#!/usr/bin/perl
#
# $Revision: 1.4 $
# Copyright 2013 Teleflora
#
# Revert to a set of edir *.tel files.
#

use strict;
use warnings;
use Getopt::Long;
use POSIX;
use English qw( -no_match_vars );
use File::Basename;
use File::Spec;

our $VERSION = '1.14';

my $CVS_REVISION = '$Revision: 1.4 $';
my $PROGNAME = basename($PROGRAM_NAME);

#######################################
#########   D E F A U L T S   #########
#######################################

my $EMPTY_STR   = q{};
my $SPACE       = q{ };
my $COMMA       = q{,};
my $COLON       = q{:};
my $ATSIGN      = q{@};
my $DOT         = q{.};
my $DOTDOT      = q{..};
my $SLASH       = q{/};
my $DASH        = q{-};
my $EQUALS      = q{=};
my $SINGLEQUOTE = q{'};

my $DEF_DAISY_TOPDIR           = '/d';
my $DAISY_DIR_NAME             = 'daisy';
my $DAISY_BIN_DIR_NAME         = 'bin';
my $DAISY_CONFIG_DIR_NAME      = 'config';
my $DAISY_LOG_DIR_NAME         = 'log';
my $DAISY_BACKUP_DIR_NAME      = 'backups';
my $DEF_DAISY_DIR              = File::Spec->catdir($DEF_DAISY_TOPDIR, $DAISY_DIR_NAME);
my $DEF_DAISY_BIN_DIR          = File::Spec->catdir($DEF_DAISY_DIR, $DAISY_BIN_DIR_NAME);
my $DEF_DAISY_CONFIG_DIR       = File::Spec->catdir($DEF_DAISY_DIR, $DAISY_CONFIG_DIR_NAME);
my $DEF_DAISY_LOG_DIR          = File::Spec->catdir($DEF_DAISY_DIR, $DAISY_LOG_DIR_NAME);
my $DEF_DAISY_BACKUP_DIR       = File::Spec->catdir($DEF_DAISY_DIR, $DAISY_BACKUP_DIR_NAME);

my $TOOLS_DIR_NAME             = 'ostools';
my $TOOLS_DIR                  = File::Spec->catdir($DEF_DAISY_TOPDIR, $TOOLS_DIR_NAME);
my $DAISY_FLORIST_DIR_NAME     = 'edirectories';
my $DAISY_FLORIST_DIR          = File::Spec->catdir($DEF_DAISY_TOPDIR, $DAISY_FLORIST_DIR_NAME);

my $FTD_FILES_ARCHIVE_NAME     = 'FTD_Files.tar';

############################################
#########   E X I T  S T A T U S   #########
############################################

my $EXIT_OK                 = 0;
my $EXIT_COMMAND_LINE       = 1;
my $EXIT_MUST_BE_ROOT       = 2;
my $EXIT_REVERT_DATE_FORMAT = 10;
my $EXIT_REVERT_UNAVAILABLE = 11;
my $EXIT_REVERT_ERROR       = 12;
my $EXIT_REVERT_NEW_FTD     = 13;

#####################################
#########   G L O B A L S   #########
#####################################

my $DAISY            = 1;
my $LOGFILE_NAME     = 'edir_revert.log';
my $LOGFILE          = File::Spec->catdir($DEF_DAISY_LOG_DIR, $LOGFILE_NAME);;
my $LOG_STDERR       = 0;
my $DEBUG            = 0;
my $TIMESTAMP        = strftime("%Y%m%d%H%M%S", localtime());

##############################################
#########   C O M M A N D  L I N E   #########
##############################################

my $REVISION = 0;
my $HELP = 0;
my $REVERT_PREVIOUS = 0;
my $REVERT_DATE = $EMPTY_STR;
my $REVERT_NEW_FTD = 0;

GetOptions(
	"version" => \$REVISION,
	"help" => \$HELP,
	"revert-previous" => \$REVERT_PREVIOUS,
	"revert-date=s" => \$REVERT_DATE,
	"revert-new-ftd" => \$REVERT_NEW_FTD,
) || die "Error: invalid command line option, exiting...\n";


# --version
if ($REVISION) {
    print("$PROGNAME $CVS_REVISION\n");
    exit($EXIT_OK);
}

# --help
if ($HELP) {
    print("$PROGNAME $CVS_REVISION\n");
    usage();
    exit($EXIT_OK);
}

#
# Don't accept any command line args... usage() and exit.
#
if ($#ARGV >= 0) {
    usage();
    exit($EXIT_COMMAND_LINE);
}

my $option_count = 0;
if ($REVERT_PREVIOUS) {
   $option_count++;
}
if ($REVERT_DATE ne $EMPTY_STR) {
   $option_count++;
}
if ($REVERT_NEW_FTD) {
   $option_count++;
}
if ($option_count > 1) {
    usage();
    exit($EXIT_COMMAND_LINE);
}

#
# Gotta be root or exit.
#
if ($EUID != 0) {
    usage();
    logerror("root privileges required to run this script");
    return($EXIT_MUST_BE_ROOT);
}

#
# Kick off the real action.
#
my $exit_status = main();

#
# Done - inform system what happened.
#
exit($exit_status);


###############################
#########   M A I N   #########
###############################

sub main
{
    my $rc = $EXIT_OK;
    my $ml = '[main]';

    loginfo("$PROGNAME $CVS_REVISION");

    my $revert_to_date = $EMPTY_STR;
    if ($REVERT_PREVIOUS) {
	$revert_to_date = get_previous_revert_date();
	if ($revert_to_date) {
	    if (edir_revert($revert_to_date)) {
		loginfo("$ml *.tel file set reverted to previous: $revert_to_date");
	    }
	    else {
		logerror("$ml could not revert to previous *.tel file set: $revert_to_date");
		$rc = $EXIT_REVERT_ERROR;
	    }
	}
	else {
	    logerror("$ml there are no previous *.tel file sets to revert to");
	    $rc = $EXIT_REVERT_UNAVAILABLE;
	}
    }
    elsif ($REVERT_DATE) {
	# check format
	if ($REVERT_DATE =~ /\d{14}/x) {
	    $revert_to_date = get_specific_revert_date($REVERT_DATE);
	    if ($revert_to_date) {
		if (edir_revert($revert_to_date)) {
		    loginfo("$ml *.tel file set reverted to: $revert_to_date");
		}
		else {
		    logerror("$ml could not revert *.tel file set to: $revert_to_date");
		    $rc = $EXIT_REVERT_ERROR;
		}
	    }
	    else {
		logerror("$ml there is no *.tel file set with date: $REVERT_DATE");
		$rc = $EXIT_REVERT_UNAVAILABLE;
	    }
	}
	else {
	    logerror("$ml revert date format: YYYYMMDDhhmmss");
	    $rc = $EXIT_REVERT_DATE_FORMAT;
	}
    }
    elsif ($REVERT_NEW_FTD) {
	if (!edir_revert_new_ftd_files()) {
	    $rc = $EXIT_REVERT_NEW_FTD;
	}
    }

    return($rc);
}


#############################################
#########   A P P L I C A T I O N   #########
#############################################

#
# revert the to a specified *.tel file set in all daisy dirs.
#
# returns
#   1 if success
#   0 on error
#
sub edir_revert
{
    my ($revert_to_date) = @_;

    my $rc = 1;

    my @db_dirs = get_daisy_db_dirs();
    foreach my $db_dir (@db_dirs) {
	if (!er_revert_db_dir($db_dir, $revert_to_date)) {
	    $rc = 0;
	}
    }

    return($rc);
}


#
# For the given daisy db dir and revert to date, revert that db dir
# to that tel set date.
#
# Returns
#   1 on success
#   0 if not
#
sub er_revert_db_dir
{
    my ($db_dir, $revert_to_date) = @_;

    my $rc = 1;
    my $ml = '[revert_db_dir]';

    my $tel_set_dir = File::Spec->catdir($db_dir, $DAISY_BACKUP_DIR_NAME);
    my $tel_set_control_file_name = 'control.tel.prepatch.' . $revert_to_date;
    my $tel_set_control_path = File::Spec->catdir($tel_set_dir, $tel_set_control_file_name);
    if (-f $tel_set_control_path) {
	if (er_restore_tel_set($db_dir, $revert_to_date)) {
	    loginfo("$ml *.tel files $revert_to_date reverted for Daisy db: $db_dir");
	}
	else {
	    logerror("$ml *.tel files $revert_to_date revert failed for: $db_dir");
	    $rc = 0;
	}
    }
    else {
	logerror("$ml *.tel set $revert_to_date does not exist for: $db_dir");
	$rc = 0;
    }

    return($rc);
}


sub er_restore_tel_set
{
    my ($db_dir, $revert_to_date) = @_;

    my $rc = 1;
    my $ml = '[restore_tel_set]';

    my $tel_set_dir = File::Spec->catdir($db_dir, $DAISY_BACKUP_DIR_NAME);
    my $tel_set_files = "/*.tel.prepatch." . $revert_to_date;
    my $tel_set_glob_expr = $tel_set_dir . $tel_set_files;

    my @tel_set_paths = glob($tel_set_glob_expr);

    if (scalar(@tel_set_paths)) {
	foreach my $src_path (@tel_set_paths) {
	    my $src_filename = basename($src_path);
	    if ($src_filename =~ /(.*)\.prepatch\.$revert_to_date/x) {
		my $dst_filename = $1;
		system("cp $src_path $db_dir/$dst_filename");
		system("chown tfsupport:daisy $db_dir/$dst_filename");
		system("chmod 660 $db_dir/$dst_filename");
	    }
	}
	loginfo("$ml *.tel files restored from: $tel_set_dir");
    }
    else {
	logerror("$ml *.tel files do not exist in: $tel_set_dir");
	$rc = 0;
    }

    return($rc);
}


#
# revert the new FTD files to each Daisy dir
#
# returns
#   1 if success
#   0 on error
#
sub edir_revert_new_ftd_files
{
    my $rc = 1;
    my $ml = '[revert_new_ftd_files]';

    my @db_dirs = get_daisy_db_dirs();
    foreach my $db_dir (@db_dirs) {
	if (er_revert_ftd_files_archive($db_dir)) {
	    loginfo("$ml FTD files reverted to: $db_dir");
	}
	else {
	    logerror("$ml could not revert FTD files to: $db_dir");
	    $rc = 0;
	}
    }

    return($rc);
}


sub er_revert_ftd_files_archive
{
    my ($db_dir) = @_;

    my $rc = 1;
    my $ml = '[revert_ftd_files_archive]';

    my $ftd_files_archive = er_pathto_ftd_files_archive();
    if (-e $ftd_files_archive) {
	my $exit_status = mysystem("tar -C $db_dir -xf $ftd_files_archive");
	if ($exit_status == 0) {
	    loginfo("$ml FTD files extracted from tar archive: $ftd_files_archive");
	}
	else {
	    logerror("$ml could not extract FTD files from tar archive $ftd_files_archive");
	    $rc = 0;
	}
    }
    else {
	logerror("$ml tar archive of FTD files does not exist: $ftd_files_archive");
	$rc = 0;
    }

    return($rc);
}


#########################################
#########   F U N C T I O N S   #########
#########################################


sub usage
{
    print("Usage:\n");
    print("$PROGNAME --version\n");
    print("$PROGNAME --help\n");
    print("$PROGNAME (options)\n");
    print("\n");

    print("Options:\n");
    print("--revert-previous\n");
    print("--revert-date=YYYYMMDDhhmmss\n");
    print("--revert-new-ftd\n");
    print("\n");

    return(1);
}


sub loginfo
{
    my ($message) = @_;

    chomp($message);
    $message = "<I> $message";
    logit($message);

    return(1);
}


sub logerror
{
    my ($message) = @_;

    chomp($message);
    $message = "<E> $message";
    logit($message);

    return(1);
}


sub logit
{
    my ($message) = @_;

    my $logtime = strftime("%Y-%m-%d %H:%M:%S", localtime());

    chomp($message);
    if (open(my $logfh, '>>', $LOGFILE)) {
	print($logfh "$logtime ($PROGNAME-$PID) $message\n");
	close($logfh);
    }
    else {
	print "$logtime ($PROGNAME-$PID) $message\n";
    }

    return(1);
}


#
# Subroutine to run system commands.  This logs an entry.
#
sub mysystem
{
    my ($command) = @_;

    my $returnval = $EMPTY_STR;
    my $log_file = $LOGFILE;

    my $stderr_dest = ($LOG_STDERR) ? "$log_file" : "/dev/null";

    system("$command >> $log_file 2>> $stderr_dest");
    $returnval = WEXITSTATUS($CHILD_ERROR);
    loginfo("$command Returned: $returnval");

    return ($returnval)
}


sub er_pathto_ftd_files_archive
{
    my $ml = '[pathto_ftd_files_archive]';

    my $ftd_files_archive = $EMPTY_STR;
    my $ftd_files_archive_re = 'FTD_Files_\d{14}\.tar';

    my $cmd = "/bin/ls $DAISY_FLORIST_DIR";
    if (open(my $pipe, '-|', $cmd)) {
	while (<$pipe>) {
	    chomp;
	    if (/$ftd_files_archive_re/) {
		$ftd_files_archive = $_;
		last;
	    }
	}
	close($pipe) or warn "$ml could not close pipe to command: $cmd ($OS_ERROR)\n";;
    }

    return(File::Spec->catdir($DAISY_FLORIST_DIR, $ftd_files_archive));
}


#
# Return a list of verified POS dirs.  For Daisy, remember, there can be
# more than one daisy database dir!
#
sub get_daisy_db_dirs
{
    my @daisy_dirs = ();
    my $daisy_top_dir = '/d';

    # Daisy POS can have one or more database dirs
    my @daisy_db_dirs = glob("$daisy_top_dir/*");
    for my $daisy_db_dir (@daisy_db_dirs) {

	# must be a directory
	next if (! -d $daisy_db_dir);

	# skip old daisy dirs
	next if ($daisy_db_dir =~ /.+-\d{12}$/x);

	# must contain the magic files
	next if (! -e "$daisy_db_dir/flordat.tel");
	next if (! -e "$daisy_db_dir/control.dsy");

	# must be daisy 8.0+
	next if (! -d "$daisy_db_dir/bin");

	push(@daisy_dirs, $daisy_db_dir);
    }

    return(@daisy_dirs);
}


#
# Count the number of tel sets in a given "backups" dir.
# Assumption: There is one "control.tel.prepatch.$timestamp" file
# per set of backed up tel files, so just count them to determine
# the number of backed up tel sets.
#
# Returns
#   number of sets
#
sub get_tel_set_count
{
    my ($db_dir) = @_;

    my @control_files = glob("$db_dir/backups/control.tel.prepatch.*");

    my $tel_set_count = scalar(@control_files);

    return($tel_set_count);
}


#
# Get a list of the tel set dates from the default Daisy database dir.
#
# Returns the number of tel sets
#
sub get_tel_set_dates
{
    my ($backup_dir, $timestamps) = @_;

    @{$timestamps} = ();

    my @control_files = glob("$backup_dir/control.tel.prepatch.*");
    if (@control_files) {
	# make a list of timestamps from each tel set
	foreach my $filename (@control_files) {
	    if ($filename =~ /.*\.tel\.prepatch\.(\d{14})/x) {
		push(@{$timestamps}, $1);
	    }
	}
    }

    return(scalar(@{$timestamps}));
}


sub get_previous_revert_date
{
    my $revert_to_date = $EMPTY_STR;
    my @timestamps = ();

    my $backup_dir = $DEF_DAISY_BACKUP_DIR;
    get_tel_set_dates($backup_dir, \@timestamps);

    # now sort them and select the last one in the list
    if (@timestamps) {
	my @timestamps_sorted = sort(@timestamps);
	$revert_to_date = $timestamps_sorted[-1];
    }

    return($revert_to_date);
}


sub get_specific_revert_date
{
    my ($revert_to_date) = @_;

    my @timestamps = ();

    my $backup_dir = $DEF_DAISY_BACKUP_DIR;
    get_tel_set_dates($backup_dir, \@timestamps);

    if (! grep {$revert_to_date} @timestamps) {
	$revert_to_date = $EMPTY_STR;
    }

    return($revert_to_date);
}



#################################################
#########   D O C U M E N T A T I O N   #########
#################################################


__END__

=pod

=head1 NAME

I<edir_revert> - script to revert to previous florist directory files

=head1 VERSION

This documentation refers to version: $Revision: 1.4 $



=head1 OPTIONS

=over 4

=item B<--version>

Output the version number of the script and exit.

=item B<--help>

Output a short help message and exit.

=item B<--revert-previous>

Revert to the most previous set of "*tel files".

=item B<--revert-date=s>

Revert to the set of "*tel files" specified by date.
The date format is "YYYYMMDDhhmmss".

=item B<--revert-new-ftd>

Revert to the new set of FTD files.

=back


=head1 DESCRIPTION

The I<edir_revert.pl> script may be used to revert to a previous
set of edir files if, for example, the newly generated files have an issue or
if the customer complains that they should be allowed to have the FTD files.

Upon execution, the script first verifies several aspects of the
environment that are required for successful completion.
If any of these checks fails, an error is written to the log file, and
the script exits with an appropriate exit status.

The following steps are performed in order:

=over 4

=item 1.

The script verifies that only one of B<--revert-previous>,
B<--revert-date=s>, or B<--revert-new-ftd> are specified on the command line.

=item 2.

A check of the effective user id is made:
the script must be running under the root account or under sudo.

=item 3.

If the B<--revert-previous> command line option was specified,
the F</d/daisy/backups> directory is searched for the date that
the last set of "*.tel files" were installed.
This date is then used to revert each of the Daisy database directories
to the set of "*.tel files" corresponding to that date.

If the B<--revert-date> command line option was specified,
the given date is searched for rather than the most recent date.

if the B<--revert-new-ftd> command line option was specifed and
there is a tar archive of saved FTD files located in F</d/edirectories>,
then they are restored to each of any existing Daisy database directories.

=back


=head1 Command Line Options

Only one of B<--revert-previous>, B<--revert-date=s> or B<--revert-new-ftd>
may appear on the command line at the same time.
If none appear, then nothing is performed.

The date format for the B<--revert-date=s> option is "YYYYMMDDhhmmss" where
"YYYY" is the year, "MM" is the month, "DD" is the day, "hh" is the hours,
"mm" is the minutes, and "ss" is the seconds.


=head1 EXAMPLES

To revert to the previous set of "*.tel files", simply enter:

$ sudo perl /d/daisy/bin/edir_revert.pl --revert-previous

To restore the removed FTD files, enter:

$ sudo perl /d/daisy/bin/edir_revert.pl --revert-new-ftd

Review the log file for details about what actually happened.


=head1 FILES

=over 4

=item F</d/daisy>

This is the Daisy default database directory and it must exist.

=item F</d/daisy/backups>

The directory where previous sets of "*.tel files" are stored.

=item F</d/daisy/log/edir_revert.log>

Log file.

=item F</d/edirectories/FTD_Files_YYYYMMDDhhmmss.tar>

When applying a static florist directory update,
if the Daisy shop is not a FTD subsriber,
the new FTD files are saved to this file,
which is a tar archive.
The string "YYYYMMDDhhmmss" is a date stamp which represents
the start of execution of the update procedure.

=back


=head1 DIAGNOSTICS

=over 4

=item Exit status 0 ($EXIT_OK)

Successful completion.

=item Exit status 1 ($EXIT_COMMAND_LINE)

In general, there was an issue with the syntax of the command line.

=item Exit status 2 ($EXIT_MUST_BE_ROOT)

The user must be root or running under "sudo".

=item Exit status 10 ($EXIT_REVERT_DATE_FORMAT)

The specified revert date must be of format: YYYYMMDDhhmmss

=item Exit status 11 ($EXIT_REVERT_UNAVAILABLE)

The specified date to revert to did not exist.

=item Exit status 12 ($EXIT_REVERT_ERROR)

There was an error restoring the selected tel set back
to a daisy database dir.

=item Exit status 13 ($EXIT_REVERT_NEW_FTD)

Could not revert to new FTD files.

=back


=head1 SEE ALSO

I<edir_installbase.pl>, I<edir_update.pl>, I<GetDeltas(1)>, I<maketel(1)>


=cut
