#!/usr/bin/perl
#
# $Revision: 1.14 $
# Copyright 2013 Teleflora
#
# Update edir by fetching deltas and generating new *.tel files.
#

use strict;
use warnings;
use Getopt::Long;
use POSIX;
use English '-no_match_vars';
use File::Basename;
use Carp;

my $CVS_REVISION = '$Revision: 1.14 $';
my $PROGNAME = basename($0);

#######################################
#########   D E F A U L T S   #########
#######################################

my $DAISY_FILESYS_DEF           = "/d";
my $DAISY_DIR_DEF               = $DAISY_FILESYS_DEF . "/daisy";
my $EDIR_BASEDIR_DEF            = $DAISY_FILESYS_DEF . "/edirectories";
my $EDIR_UPDATE_CONF_DEF        = $DAISY_DIR_DEF     . "/config/edir_update.conf";
my $EDIR_LOCKFILE_DEF           = $DAISY_DIR_DEF     . "/config/edir_update.lock";
my $EDIR_PIDFILE_DEF            = $DAISY_DIR_DEF     . "/config/edir_update.pid";
my $LOGFILE_DEF                 = $DAISY_DIR_DEF     . "/log/edir_update.log";
my $SUMMARY_LOGFILE_DEF         = $DAISY_DIR_DEF     . "/log/Delta_Stats.log";
my $EDIR_BOOK_DEF               = "may";
my $EDIR_YEAR_DEF               = "2015";
my $EDIR_DATEDIR_DEF            = $EDIR_BASEDIR_DEF . '/' . $EDIR_YEAR_DEF . $EDIR_BOOK_DEF;
my $EDIR_GETDELTAS_NAME         = "GetDeltas";  # name of external program 1
my $EDIR_GETDELTAS_TIMEOUT_DEF  = 720;          # default timeout in seconds (12 minutes)
my $EDIR_MAKETEL_NAME           = "maketel";    # name of external program 2
my $EDIR_MAKETEL_TIMEOUT_DEF    = 720;          # default timeout in seconds (12 minutes)
my $EDIR_MAKETEL_HOUR_DEF       = 2;            # only run during this hour
my $EDIR_MAX_TEL_SETS_DEF       = 3;            # max nr of tel sets to keep
my $MAX_RANDOM_MINUTES_DEF      = 60;           # max nr of minutes to delay

############################################
#########   E X I T  S T A T U S   #########
############################################

my $EXIT_OK                     = 0;
my $EXIT_ERROR                  = -1;
my $EXIT_COMMAND_LINE           = 1;
#my $EXIT_MUST_BE_ROOT           = 2;
my $EXIT_DISK_SPACE             = 3;
my $EXIT_EDIR_TREE              = 4;
my $EXIT_SUMMARY_LOGFILE        = 5;
my $EXIT_CMD_MISSING            = 6;
my $EXIT_CMD_PERMS              = 7;
my $EXIT_GETDELTAS_TIMEOUT      = 8;
my $EXIT_GETDELTAS_EXCEPTION    = 9;
my $EXIT_GETDELTAS_ERROR        = 10;
my $EXIT_MAKETEL_TIMEOUT        = 11;
my $EXIT_MAKETEL_EXCEPTION      = 12;
my $EXIT_MAKETEL_ERROR          = 13;
my $EXIT_NO_NEW_DELTAS          = 14;
my $EXIT_PATCH_DIR              = 15;
my $EXIT_PATCH_BINDIR           = 16;
my $EXIT_MAKE_PATCH             = 17;
my $EXIT_PATCH_NOT_FOUND        = 18;
my $EXIT_APPLY_PATCH            = 19;
my $EXIT_LOCKFILE_MAKE_ERR      = 20;
my $EXIT_LOCKFILE_RM_ERR        = 21;
my $EXIT_GETDELTAS_CONF         = 22;

#####################################
#########   G L O B A L S   #########
#####################################

my $EMPTY                       = q{};
my $DAISY                       = 1;
my $DAISY_DIR                   = $DAISY_DIR_DEF;
my $LOGFILE                     = $LOGFILE_DEF;
my $EDIR_LOCKFILE               = $EDIR_LOCKFILE_DEF;
my $EDIR_PIDFILE                = $EDIR_PIDFILE_DEF;
my $DEBUG                       = 0;
my $TIMESTAMP                   = strftime("%Y%m%d%H%M%S", localtime());
my $ALTROOT                     = $EMPTY;         # alternate root for testing
my $ALTOS                       = $EMPTY;         # alternate os for testing

# 
# All date formats used in the log file will be in the format of:
#   "<dayname>, <monthname> <monthday>, <year> at <hour>:<min>".
# 
# All lines are a string, followed by a ": ", and a value
#
#          string                          value
# line 1:  "CURRENT UPDATE STATUS"         "SUCCESSFUL" | "FAILURE"
# line 2:  "CURRENT BOOK NAME"             "JAN" | "MAY" | "SEP"
# line 3:  "CURRENT BOOK YEAR"             <yyyy>
# line 4:  "LAST DELTA REQUEST DATE"       <date>
# line 5:  "LAST SUCCESSFUL DIR UPDATE     <date>
# line 6:  "LAST SUCCESSFUL PRODUCT UPDATE <date>
# line 7:  "LAST SUCCESSFUL DIR SEQ NR"    <nnnnnnnnnnnnnnnn>
# line 8:  "LAST SUCCESSFUL PROD SEQ NR"   <nnnnnnnnnnnnnnnn>
#
# If value of line 1 is "FAILURE", then there will be several more lines.
#
# line 9:  "FAILURE DATE"                  <date>
# line 10: "FAILURE DESCRIPTION"           <text>
# line 11: "FAILURE EXIT STATUS GETDELTAS" <text>
# line 12: "FAILURE EXIT STATUS MAKETEL"   <text>
#
my @SummaryLogFileRecs = (
    {rec_name => 'Status',            rec_desc => 'CURRENT UPDATE STATUS',         rec_value => $EMPTY},
    {rec_name => 'Book',              rec_desc => 'CURRENT BOOK NAME',             rec_value => $EMPTY},
    {rec_name => 'Year',              rec_desc => 'CURRENT BOOK YEAR',             rec_value => $EMPTY},
    {rec_name => 'DeltaReqDate',      rec_desc => 'LAST DELTA REQUEST DATE',       rec_value => $EMPTY},
    {rec_name => 'DirUpdate',         rec_desc => 'LAST SUCCESSFUL DIR UPDATE',    rec_value => $EMPTY},
    {rec_name => 'ProdUpdate',        rec_desc => 'LAST SUCCESSFUL PROD UPDATE',   rec_value => $EMPTY},
    {rec_name => 'DirSeqNr',          rec_desc => 'LAST SUCCESSFUL DIR SEQ NR',    rec_value => $EMPTY},
    {rec_name => 'ProdSeqNr',         rec_desc => 'LAST SUCCESSFUL PROD SEQ NR',   rec_value => $EMPTY},
    {rec_name => 'FailDate',          rec_desc => 'FAILURE DATE',                  rec_value => $EMPTY},
    {rec_name => 'FailDesc',          rec_desc => 'FAILURE DESCRIPTION',           rec_value => $EMPTY},
    {rec_name => 'FailGetDeltasExit', rec_desc => 'FAILURE EXIT STATUS GETDELTAS', rec_value => $EMPTY},
    {rec_name => 'FailMaketelExit',   rec_desc => 'FAILURE EXIT STATUS MAKETEL',   rec_value => $EMPTY},
);
my $SummaryLogFileRecCount = scalar(@SummaryLogFileRecs);

##############################################
#########   C O M M A N D  L I N E   #########
##############################################

my $VERSION = 0;
my $HELP = 0;
my $VERBOSE = 0;
my $EDIR_UPDATE_CONF = $EDIR_UPDATE_CONF_DEF;
my $EDIR_BASEDIR = $EDIR_BASEDIR_DEF;
my $EDIR_BOOK = $EDIR_BOOK_DEF;
my $EDIR_YEAR = $EDIR_YEAR_DEF;
my $EDIR_GETDELTAS_TIMEOUT = $EDIR_GETDELTAS_TIMEOUT_DEF;
my $EDIR_MAKETEL_TIMEOUT = $EDIR_MAKETEL_TIMEOUT_DEF;
my $EDIR_FORCE_MAKETEL = 0;
my $EDIR_SKIP_UPDATE = 0;
my $EDIR_LOCKFILE_RM = 0;
my $EDIR_REPORT_CONFIGFILE = 0;
my $EDIR_GEN_SUMMARY_LOG = 0;	# temporary default to no
# three possibilities:
#   if 0, no delay
#   if < 0, random delay
#   if > 0, delay $EDIR_DELAY seconds
my $EDIR_DELAY = -1;
my $EDIR_MAX_TEL_SETS = $EDIR_MAX_TEL_SETS_DEF;

GetOptions(
	"version" => \$VERSION,
	"help" => \$HELP,
	"verbose" => \$VERBOSE,
	"altroot=s" => \$ALTROOT,
	"altos=s" => \$ALTOS,
	"edir-config=s" => \$EDIR_UPDATE_CONF,
	"report-config" => \$EDIR_REPORT_CONFIGFILE,
	"basedir|edir-basedir=s" => \$EDIR_BASEDIR,
	"book|edir-book=s" => \$EDIR_BOOK,
	"year|edir-year=s" => \$EDIR_YEAR,
	"edir-getdeltas-timeout=s" => \$EDIR_GETDELTAS_TIMEOUT,
	"edir-maketel-timeout=s" => \$EDIR_MAKETEL_TIMEOUT,
	"force-maketel" => \$EDIR_FORCE_MAKETEL,
	"skip-update" => \$EDIR_SKIP_UPDATE,
	"lockfile-remove" => \$EDIR_LOCKFILE_RM,
	"summary-logfile!" => \$EDIR_GEN_SUMMARY_LOG,
	"delay=s" => \$EDIR_DELAY,
	"savedtelsets=s" => \$EDIR_MAX_TEL_SETS,
) || die "Error: invalid command line option, exiting...\n";

my $EDIR_BINDIR_NAME     = "bin";
my $EDIR_CONFDIR_NAME    = "config";
my $EDIR_DELTADIR_NAME   = "deltas";
my $EDIR_TELDIR_NAME     = "tel";
my $EDIR_PATCHDIR_NAME   = "patchdir";
my $EDIR_PATCH_TYPE      = "tel";
my $EDIR_PATCH_EXTENSION = ".patch";
my $EDIR_PATCH_SUFFIX    = '_' . $EDIR_PATCH_TYPE . $EDIR_PATCH_EXTENSION;


#
# read the config file
#
eu_read_configfile($EDIR_UPDATE_CONF);

#
# sanity check of values from command line and config file.
#
if ($EDIR_MAX_TEL_SETS < 0) {
    $EDIR_MAX_TEL_SETS = $EDIR_MAX_TEL_SETS_DEF;
}

#
# Global values which depend on command line options or
# config file.
#
my $EDIR_DATEDIR         = $EDIR_BASEDIR . '/' . $EDIR_YEAR . $EDIR_BOOK;

my $EDIR_BINDIR          = $EDIR_DATEDIR . '/' . $EDIR_BINDIR_NAME;
my $EDIR_CONFDIR         = $EDIR_DATEDIR . '/' . $EDIR_CONFDIR_NAME;
my $EDIR_DELTADIR        = $EDIR_DATEDIR . '/' . $EDIR_DELTADIR_NAME;
my $EDIR_TELDIR          = $EDIR_DATEDIR . '/' . $EDIR_TELDIR_NAME;
my $EDIR_PATCHDIR        = $EDIR_DATEDIR . '/' . $EDIR_PATCHDIR_NAME;
my $EDIR_PATCH_NAME      = $EDIR_YEAR . $EDIR_BOOK . $EDIR_PATCH_SUFFIX;

my $EDIR_GENERATED       = $EDIR_CONFDIR . "/initial_generation_done";
my $EDIR_DELTA_COUNTFILE = $EDIR_CONFDIR . "/edir_update_delta_count.txt";
my $EDIR_DELTA_LAST_REQ  = $EDIR_CONFDIR . "/edir_update_delta_last_request.txt";
my $EDIR_GETDELTAS_CONF  = $EDIR_CONFDIR . '/' . $EDIR_GETDELTAS_NAME . ".conf";
my $EDIR_MAKETEL_CONF    = $EDIR_CONFDIR . '/' . $EDIR_MAKETEL_NAME . ".conf";

my $EDIR_GETDELTAS_CMD   = $EDIR_BINDIR . '/' . $EDIR_GETDELTAS_NAME;
my $EDIR_MAKETEL_CMD     = $EDIR_BINDIR . '/' . $EDIR_MAKETEL_NAME;

# --version
if ($VERSION) {
	print("$PROGNAME $CVS_REVISION\n");
	exit($EXIT_OK);
}

# --help
if ($HELP) {
	print("$PROGNAME $CVS_REVISION\n");
	usage();
	exit($EXIT_OK);
}

#
# Don't accept any command line args... usage() and exit.
#
if ($#ARGV >= 0) {
	usage();
	exit($EXIT_COMMAND_LINE);
}

#
# Kick off the real action.
#
my $exit_status = main();

#
# Generate summary log file.
#
if ($EDIR_GEN_SUMMARY_LOG) {
    if (eu_summary_logfile_generate()) {
	loginfo("summary log file generated: $SUMMARY_LOGFILE_DEF");
    }
    else {
	logerror("could not generate summary log file: $SUMMARY_LOGFILE_DEF");
    }
}

#
# Done - inform system what happened.
#
exit($exit_status);


###############################
#########   M A I N   #########
###############################

sub main
{
    loginfo("$PROGNAME $CVS_REVISION");

    eu_summary_logfile_set_rec_value_by_rec_name('Book', $EDIR_BOOK);
    eu_summary_logfile_set_rec_value_by_rec_name('Year', $EDIR_YEAR);

    #
    # Verify file system with the daisy database dir is not 100% full.
    #
    if (get_diskused($DAISY_FILESYS_DEF) == 100) {
	my $msg = "Daisy filesystem is 100% full: $DAISY_FILESYS_DEF";
	logerror($msg);
	eu_summary_logfile_note_failure($msg);
	return($EXIT_DISK_SPACE);
    }

    #
    # Verify edir tree is present.
    #
    unless (-d $EDIR_BASEDIR) {
	my $msg = "edir directory does not exist: $EDIR_BASEDIR";
	logerror($msg);
	eu_summary_logfile_note_failure($msg);
	return($EXIT_EDIR_TREE);
    }

    #
    # If specified, remove the lockfile before proceeding.
    #
    if ($EDIR_LOCKFILE_RM) {
	if (eu_lockfile_rm()) {
	    loginfo("lockfile removed");
	}
	else {
	    return($EXIT_LOCKFILE_RM_ERR);
	}
    }

    #
    # wait a random amount of time up to 59 minutes... this is so
    # not all Daisy systems in the same time zone will try to connect
    # to the Delta Force web server at the same time.
    #
    if ($EDIR_DELAY < 0) {
	my $random_minutes = int(rand($MAX_RANDOM_MINUTES_DEF));
	if ( ($random_minutes < 0) || ($random_minutes >= $MAX_RANDOM_MINUTES_DEF) ) {
	    logerror("rand($MAX_RANDOM_MINUTES_DEF) returns number out of range: $random_minutes");
	    $random_minutes = ($MAX_RANDOM_MINUTES_DEF / 2);
	}
	loginfo("delaying random number of minutes: $random_minutes");
	my $random_seconds = $random_minutes * 60;
	sleep($random_seconds);
	loginfo("delay complete");
    }
    elsif ($EDIR_DELAY == 0) {
	loginfo("not delaying");
    }
    else {
	loginfo("user specified delay (seconds): $EDIR_DELAY");
	sleep($EDIR_DELAY);
	loginfo("delay complete");
    }

    #
    # attempt to make a lockfile so other instances of script can not
    # start while this one is running.  Block signals so that the script
    # can cleanup lock file.
    #
    local $SIG{'TSTP'} = 'IGNORE';
    local $SIG{'INT'} = 'IGNORE';
    unless (eu_lockfile_make()) {
	my $msg = "lockfile make error";
	logerror($msg);
	eu_summary_logfile_note_failure($msg);
	local $SIG{'TSTP'} = 'DEFAULT';
	local $SIG{'INT'} = 'DEFAULT';

	return($EXIT_LOCKFILE_MAKE_ERR);
    }

    #
    # let others know what process is running
    #
    eu_pidfile_make();

    #
    # Finally can attempt task at hand.
    #
    loginfo("Updating edir");
    my $rc = edir_update($EDIR_BASEDIR, $EDIR_BOOK, $EDIR_YEAR);
    if ($rc == $EXIT_OK) {
	loginfo("edir update successful");
	eu_summary_logfile_note_success();
    }
    else {
	my $msg = "update of edir returned status: $rc";
	logerror($msg);
	eu_summary_logfile_note_failure($msg);
    }

    #
    # clean up
    #
    eu_pidfile_rm();
    eu_lockfile_rm();
    local $SIG{'TSTP'} = 'DEFAULT';
    local $SIG{'INT'} = 'DEFAULT';

    loginfo("$PROGNAME exiting with status $rc");

    return($rc);
}


#
# Read config file.
#
# Returns
#   1 on success
#   0 on error
#
sub eu_read_configfile
{
    my ($configfile) = @_;

    if ($configfile eq $EMPTY) {
	logerror("can't happen: config file path value is the empty string");
	return(0);
    }
    unless (-e $configfile) {
	loginfo("config file does not exist: $configfile");
	return(1);
    }

    loginfo("reading config file: $configfile");

    if (open(my $cfh, "<", $configfile)) {
	while (<$cfh>) {
	    my $normalized_line = eu_parse_configfile_line($_);
	    if ($normalized_line && $EDIR_REPORT_CONFIGFILE) {
		print "$normalized_line\n";
	    }
	}
	close($cfh);
    }
    else {
	logerror("can't open config file: $configfile");
	return(0);
    }

    return(1);
}


sub eu_parse_configfile_line
{
    my ($line) = @_;

    my $normalized_line = $EMPTY;

    # report-config=0
    if ($line =~ /^(\s*)(report-config)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	if ($6 =~ /[YyTt1]/x) {
	    $EDIR_REPORT_CONFIGFILE = 1;
	    $normalized_line = "report-config=1";
	}
	elsif ($6 =~ /[YnFf0]/x) {
	    $EDIR_REPORT_CONFIGFILE = 0;
	    $normalized_line = "report-config=0";
	}
    }

    # edir-basedir=string
    if ($line =~ /^(\s*)(edir-basedir)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	$EDIR_BASEDIR = $6;
	$normalized_line = "edir-basedir=$6";
    }

    # edir-book=string
    if ($line =~ /^(\s*)(edir-book)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	$EDIR_BOOK = $6;
	$normalized_line = "edir-book=$6";
    }

    # edir-year=string
    if ($line =~ /^(\s*)(edir-year)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	$EDIR_YEAR = $6;
	$normalized_line = "edir-year=$6";
    }

    # edir-getdeltas-timeout=string
    if ($line =~ /^(\s*)(edir-getdeltas-timeout)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	$EDIR_GETDELTAS_TIMEOUT = $6;
	$normalized_line = "edir-getdeltas-timeout=$6";
    }

    # edir-maketel-timeout=string
    if ($line =~ /^(\s*)(edir-maketel-timeout)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	$EDIR_MAKETEL_TIMEOUT = $6;
	$normalized_line = "edir-maketel-timeout=$6";
    }

    # force-maketel=0
    if ($line =~ /^(\s*)(force-maketel)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	if ($6 =~ /[YyTt1]/x) {
	    $EDIR_FORCE_MAKETEL = 1;
	    $normalized_line = "force-maketel=1";
	}
	elsif ($6 =~ /[YnFf0]/x) {
	    $EDIR_FORCE_MAKETEL = 0;
	    $normalized_line = "force-maketel=0";
	}
    }

    # skip-update=0
    if ($line =~ /^(\s*)(skip-update)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	if ($6 =~ /[YyTt1]/x) {
	    $EDIR_SKIP_UPDATE = 1;
	    $normalized_line = "skip-update=1";
	}
	elsif ($6 =~ /[YnFf0]/x) {
	    $EDIR_SKIP_UPDATE = 0;
	    $normalized_line = "skip-update=0";
	}
    }

    # lockfile-remove=0
    if ($line =~ /^(\s*)(lockfile-remove)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	if ($6 =~ /[YyTt1]/x) {
	    $EDIR_LOCKFILE_RM = 1;
	    $normalized_line = "lockfile-remove=1";
	}
	elsif ($6 =~ /[NnFf0]/x) {
	    $EDIR_LOCKFILE_RM = 0;
	    $normalized_line = "lockfile-remove=0";
	}
    }

    # summary-logfile=1
    if ($line =~ /^(\s*)(summary-logfile)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	if ($6 =~ /[YyTt1]/x) {
	    $EDIR_GEN_SUMMARY_LOG = 1;
	    $normalized_line = "summary-logfile=1";
	}
	elsif ($6 =~ /[NnFf0]/x) {
	    $EDIR_GEN_SUMMARY_LOG = 0;
	    $normalized_line = "summary-logfile=0";
	}
    }

    # savedtelsets=n
    if ($line =~ /^(\s*)(savedtelsets)(\s*)(=)(\s*)([[:print:]]+)$/ix) {
	$EDIR_MAX_TEL_SETS = $6;
	$normalized_line = "savedtelsets=$EDIR_MAX_TEL_SETS";
    }

    return($normalized_line);
}


#############################################
#########   A P P L I C A T I O N   #########
#############################################

sub edir_update
{
    my ($edir_basedir, $edir_book, $edir_year) = @_;

    my $edir_datedir = $edir_basedir . '/' . $edir_year . $edir_book;
    my $rc = $EXIT_OK;

    #
    # verify the external commands are present and executable
    #
    my @external_commands = (
	"$EDIR_GETDELTAS_CMD",
	"$EDIR_MAKETEL_CMD",
    );
    $rc = eu_verify_external_commands(@external_commands);
    if ($rc != $EXIT_OK) {
	return($rc);
    }

    #
    # short circuit update - useful in test environment.
    #
    if ($EDIR_SKIP_UPDATE) {
	loginfo("skip-update=TRUE, skipping update");
	return($rc);
    }

    #
    # exec the first external command, exit if not successful.
    #
    $rc = eu_exec_getdeltas($edir_basedir, $edir_year, $edir_book);
    if ($rc != $EXIT_OK) {
	return($rc);
    }
  
    #
    # count the number of delta files... if we got some more...
    # we are eligible to make the "tel" files.
    #
    unless (eu_compare_delta_file_counts($EDIR_DELTADIR, $EDIR_DELTA_COUNTFILE)) {
	my $desc = "no new delta files, no need to run maketel";
	loginfo($desc);
	eu_summary_logfile_set_rec_value_by_rec_name('FailGetDeltasExit', $desc);
	$desc = "maketel was not run";
	eu_summary_logfile_set_rec_value_by_rec_name('FailMaketelExit', $desc);

	return($EXIT_NO_NEW_DELTAS);
    }

    #
    # verify that current time is between 2am and 3am
    #
    unless ($EDIR_FORCE_MAKETEL) {
	my $hour = (localtime())[2];
	if ($hour != $EDIR_MAKETEL_HOUR_DEF) {
	    loginfo("maketel will only run during this hour: $EDIR_MAKETEL_HOUR_DEF");
	    loginfo("current hour is: $hour");

	    return($EXIT_OK);
	}
    }

    #
    # at this point, first external command has run successfully, so
    # go on to the second external command.
    #
    $rc = eu_exec_maketel($edir_basedir, $edir_year, $edir_book);
    if ($rc != $EXIT_OK) {
	return($rc);
    }

    #
    # put the new "tel" files into use.
    #
    my $edir_date = $edir_year . $edir_book;
    $rc = eu_make_edir_patch($EDIR_PATCHDIR, $EDIR_BINDIR, $EDIR_TELDIR, $edir_date);
    if ($rc == $EXIT_OK) {
	$rc = eu_apply_edir_patch($EDIR_PATCHDIR, $EDIR_PATCH_NAME, $EDIR_BINDIR);
    }
    if ($rc != $EXIT_OK) {
	return($rc);
    }

    #
    # clean up the "backups" directory in all the Daisy database dirs.
    # Only keep the last "n" set of *.tel files where "n" is a define.
    #
    unless (eu_purge_tel_sets($EDIR_MAX_TEL_SETS)) {
	logerror("can't purge old tel sets down to: $EDIR_MAX_TEL_SETS");
    }

    unless (eu_first_generation_marker_exists()) {
	eu_make_first_generation_marker();
    }

    #
    # don't forget to update the delta count file
    #
    eu_update_delta_count_file($EDIR_DELTADIR, $EDIR_DELTA_COUNTFILE);

    return($EXIT_OK);
}


#
# verify that the commands called by this script fulfill
# the requirements needed to exec them: they can be found and
# they have the appropriate perms.
#
# Returns
#   $EXIT_OK on success
#   $EXIT_CMD_PERMS if command file is not executable
#   $EXIT_CMD_MISSING if command file does not exist
#
sub eu_verify_external_commands
{
    my (@cmds) = @_;

    foreach (@cmds) {
	if (-f $_) {
	    loginfo("edir update command present: $_");
	    if (-x $_) {
		loginfo("edir update command perms correct: $_");
	    }
	    else {
		logerror("edir command file not executable: $_\n");
		return($EXIT_CMD_PERMS);
	    }
	}
	else {
	    logerror("edir command file does not exist: $_\n");
	    return($EXIT_CMD_MISSING);
	}
    }

    return($EXIT_OK);
}


#
# only one instance of the script is allowed to run at a time.
# If the lockfile exists, then there is already an instance
# of the script running.
#
# Returns
#   TRUE (1) if no other instance is running
#   FALSE (0) if another instance is already running
#
sub eu_verify_exclusive_instance
{
    if (-f $EDIR_LOCKFILE) {
	logerror("Lockfile for $PROGNAME exists, try again later: $EDIR_LOCKFILE");
	return(0);
    }

    return(1);
}


#
# attempt to make the lockfile.
#
# Returns
#   1 if successful
#   0 if not
#
# FIXME: if making the lock file fails, look for a pid file and
# if there is a process that corresponds to the pid.  If there
# is no pid file, look for a "edir_update.pl" process by name.
# if there is no currently running instance, there should not be
# a lock file or a pid file.
#
sub eu_lockfile_make
{
    unless (sysopen(LFH, $EDIR_LOCKFILE, O_EXCL|O_CREAT)) {
	logerror("exclusive open of lockfile failed, try again later: $EDIR_LOCKFILE");
	return(0);
    }
    close(LFH);

    return(1);
}


#
# remove the lockfile.
#
# Returns
#   1 if successful
#   0 if not
#
sub eu_lockfile_rm
{
    my $rc = 1;

    if (-f $EDIR_LOCKFILE) {
	if (unlink($EDIR_LOCKFILE)) {
	    loginfo("lockfile removed: $EDIR_LOCKFILE");
	}
	else {
	    logerror("lockfile unlink failed: $EDIR_LOCKFILE");
	    $rc = 0;
	}
    }
    else {
	logerror("lockfile unlink not attempted, lockfile does not exist: $EDIR_LOCKFILE");
	$rc = 0;
    }

    return(1);
}


sub eu_pidfile_make
{
    my $rc = 1;

    if (open(my $pfh, '>', $EDIR_PIDFILE)) {
	print $pfh "pid: $$\n";
	close($pfh);
	loginfo("pid ($$) writen to pidfile: $EDIR_PIDFILE");
    }
    else {
	logerror("write of pid file failed: $EDIR_PIDFILE");
	$rc = 0;
    }

    return($rc);
}


sub eu_pidfile_rm
{
    my $rc = 1;

    if (-f $EDIR_PIDFILE) {
	if (unlink($EDIR_PIDFILE)) {
	    loginfo("pidfile removed: $EDIR_PIDFILE");
	}
	else {
	    logerror("pidfile unlink failed: $EDIR_PIDFILE");
	    $rc = 0;
	}
    }
    else {
	logerror("pidfile unlink not attempted, pidfile does not exist: $EDIR_PIDFILE");
	$rc = 0;
    }

    return($rc);
}


#
# Exec the "GetDeltas" command.
#
# Returns
#   $EXIT_OK if successful run
#   non-zero exit status if not
#
sub eu_exec_getdeltas
{
    my ($edir_basedir, $edir_year, $edir_book) = @_;

    my $rc = $EXIT_OK;

    #
    # make the conf file
    #
    # we would rather use command line options like:
    #$cmd .= " -q $edir_book";
    #$cmd .= " -y $edir_year";
    #$cmd .= " -d $edir_basedir";
    #
    system("echo $edir_year > $EDIR_GETDELTAS_CONF");
    system("echo $edir_book >> $EDIR_GETDELTAS_CONF");
    system("echo 0 >> $EDIR_GETDELTAS_CONF");
    system("echo $edir_basedir >> $EDIR_GETDELTAS_CONF");
    system("chown tfsupport:dsyadmins $EDIR_GETDELTAS_CONF");
    system("chmod 0644 $EDIR_GETDELTAS_CONF");
    unless (-f $EDIR_GETDELTAS_CONF) {
	logerror("can't generate conf file: $EDIR_GETDELTAS_CONF");
	return($EXIT_GETDELTAS_CONF);
    }

    # form the command line
    my $cmd = $EDIR_GETDELTAS_CMD;
    $cmd .= " --config $EDIR_GETDELTAS_CONF";

    # call the command with a timeout
    loginfo("calling command: $cmd");
    my $rc_getdeltas = eu_system_with_timeout($cmd, $EDIR_GETDELTAS_TIMEOUT_DEF); 
    if ($rc_getdeltas == -1) {
	logerror("exiting with timeout while running: $cmd");
	$rc = $EXIT_GETDELTAS_TIMEOUT;
    }
    elsif ($rc_getdeltas == -2) {
	logerror("exiting with exception while running: $cmd");
	$rc = $EXIT_GETDELTAS_EXCEPTION;
    }
    elsif ($rc_getdeltas != 0) {
	logerror("exiting with non-zero exit status $rc_getdeltas from: $cmd");
	my $desc = eu_select_exit_status_desc($rc_getdeltas);
	eu_summary_logfile_set_rec_value_by_rec_name('FailGetDeltasExit', $desc);
	$rc = $EXIT_GETDELTAS_ERROR;
    }
    else {
	loginfo("successful exit status from: $cmd");
    }

    # this records time of last delta request
    system("touch $EDIR_DELTA_LAST_REQ");
    unless (-f $EDIR_DELTA_LAST_REQ) {
	logerror("could not record time of last delta request");
    }

    return($rc);
}


sub eu_exec_maketel
{
    my ($edir_basedir, $edir_year, $edir_book) = @_;

    my $rc = $EXIT_OK;

    my $cmd = $EDIR_MAKETEL_CMD;
    $cmd .= " -q $edir_book";
    $cmd .= " -y $edir_year";
    $cmd .= " -d $edir_basedir";
    $cmd = "cd $EDIR_TELDIR; $cmd";
    loginfo("calling command: $cmd");
    my $rc_maketel = eu_system_with_timeout($cmd, $EDIR_MAKETEL_TIMEOUT_DEF); 
    if ($rc_maketel == -1) {
	logerror("exiting with timeout while running: $cmd");
	$rc = $EXIT_MAKETEL_TIMEOUT;
    }
    elsif ($rc_maketel == -2) {
	logerror("exiting with exception while running: $cmd");
	$rc = $EXIT_MAKETEL_EXCEPTION;
    }
    elsif ($rc_maketel != 0) {
	logerror("exiting with non-zero exit status $rc from: $cmd");
	my $desc = eu_select_exit_status_desc($rc_maketel);
	eu_summary_logfile_set_rec_value_by_rec_name('FailMaketelExit', $desc);
	$rc = $EXIT_MAKETEL_ERROR;
    }
    else {
	loginfo("Successful exit status from: $cmd");
    }

    return($rc);
}

#
# Make an old style Daisy edir patch file
#
# make patch dir
# make patch bin dir
# copy postpatch.pl to patch bin dir
# copy edir program to patch dir
# copy *.tel files to patch dir
# copy us-cdifcatm.txt file to patch dir
#
# Returns
#   $EXIT_OK on success
#   $EXIT_PATCH_DIR if patchdir does not exist or can't be made
#   $EXIT_PATCH_BINDIR if bindir in patchdir does not exist or ...
#
sub eu_make_edir_patch
{
    my ($patchdir, $edir_bindir, $edir_teldir, $edir_date) = @_;

    loginfo("making patch file");

    unless (-d $patchdir) {
	system("mkdir $patchdir");
	unless (-d $patchdir) {
	    logerror("can't make patch directory: $patchdir");
	    return($EXIT_PATCH_DIR);
	}
    }
    my $patch_bindir = "$patchdir/bin";
    unless (-d $patch_bindir) {
	system("mkdir $patch_bindir");
	unless (-d $patch_bindir) {
	    logerror("can't make patch bin directory: $patch_bindir");
	    return($EXIT_PATCH_BINDIR);
	}
    }

    #
    # clean out the old "tel" files.
    #
    loginfo("cleaning out patch dir");
    system("rm -f $patchdir/*.tel");
    system("rm -f $patchdir/*.patch");
    system("rm -f $patchdir/us-cdifcatm.txt");

    #
    # copy the binaries and script to bin dir
    #
    loginfo("copying script and binaries to patch dir");
    my $srcdir = "$edir_bindir";
    system("cp $srcdir/postpatch.pl $patch_bindir");

    #
    # if this is the first time that "*.tel" files were generated
    # for this book and year, then put the new "edir" program and
    # it's corresponding screen file in the patch.
    #
    unless (eu_first_generation_marker_exists()) {
	system("cp $srcdir/edir-* $patchdir");
	system("cp $srcdir/dir_scr.scr $patchdir");
    }

    #
    # copy the "tel" files
    #
    loginfo("copying tel files to patch dir");
    $srcdir = "$edir_teldir";
    system("cp $srcdir/*.tel $patchdir");
    system("cp $srcdir/us-cdifcatm.txt $patchdir");

    #
    # make the patch file
    #
    my $cmd = "$edir_bindir/makepatch.pl";
    $cmd .= " --build $edir_date";
    $cmd .= " --name $EDIR_PATCH_TYPE";
    $cmd .= " *";
    loginfo("calling make patch script: $cmd");
    system("cd $patchdir; $cmd");
    if ($? != 0) {
	logerror("make patch cmd: $cmd");
	logerror("make patch cmd returned: $?");
	return($EXIT_MAKE_PATCH);
    }

    return($EXIT_OK);
}


sub eu_apply_edir_patch
{
    my ($patchdir, $patch_name, $edir_bindir) = @_;

    #
    # verify existence of patch file
    #
    unless (-f "$patchdir/$patch_name") {
	logerror("patch file does not exist: $patchdir/$patch_name");
	return($EXIT_PATCH_NOT_FOUND);
    }

    #
    # apply the patch file
    #
    my $cmd = "$edir_bindir/applypatch.pl";
    $cmd .= " --norestart";
    $cmd .= " --log-stderr";
    $cmd .= " --install-ftd-files";
    $cmd .= " $patch_name";
    loginfo("calling apply patch script: $cmd");
    system("cd $patchdir; $cmd");
    if ($? != 0) {
	logerror("apply patch cmd returned: $?");
	return($EXIT_APPLY_PATCH);
    }

    return($EXIT_OK);
}


sub eu_count_delta_files
{
    my ($edir_deltadir) = @_;

    my @delta_files = glob("$edir_deltadir/*.xml");

    return(scalar(@delta_files));
}


#
# Is this the first time *.tel files have been generated for
# this "year" and "book"?
#
# Returns
#   1 if first *.tel files already generated
#   0 if not
#
sub eu_first_generation_marker_exists
{
    # if this file exists, it means that the first set of *.tel files
    # has already been generated and thus the files that need to be
    # replaced have already been replaced.
    if (-f $EDIR_GENERATED) {
	return(1);
    }

    return(0);
}


#
# Make the marker that the first generation of *.tel files has
# already been done.
#
# Returns
#   1 for success
#   0 for error
#
sub eu_make_first_generation_marker
{
    system("touch $EDIR_GENERATED");
    unless (-f $EDIR_GENERATED) {
	logerror("can't make first generation already done marker file: $EDIR_GENERATED");
	return(0);
    }

    loginfo("the first generation already done marker file made: $EDIR_GENERATED");
    return(1);
}


#
# Count the number of delta files in the delta dir and compare it
# to the count saved in the delta count file.
#
# Returns
#   1 if there are new delta files
#   0 if not
#
sub eu_compare_delta_file_counts
{
    my ($edir_deltadir, $delta_countfile) = @_;

    my $rc = 0;

    my $delta_count_current = eu_count_delta_files($edir_deltadir);

    unless (-f $delta_countfile) {
	system("echo 0 > $delta_countfile");
	system("chown tfsupport:dsyadmins $delta_countfile");
	system("chmod 0644 $delta_countfile");
    }

    if (open(my $cfh, '<', $delta_countfile)) {
	my $delta_count_prev = <$cfh>;
	chomp($delta_count_prev);
	close($cfh);

	if ($delta_count_current > $delta_count_prev) {
	    $rc = 1;
	}
    }

    return($rc);
}


sub eu_update_delta_count_file
{
    my ($edir_deltadir, $delta_countfile) = @_;

    my $delta_count_current = eu_count_delta_files($edir_deltadir);

    system("echo $delta_count_current > $delta_countfile");
    system("chown tfsupport:dsyadmins $delta_countfile");
    system("chmod 0644 $delta_countfile");

    return(1);
}

#
# Purge old tel sets from "backups" directory.
#
# Returns
#   1 on success
#   0 on error
#
sub eu_purge_tel_sets
{
    my ($edir_max_tel_sets) = @_;

    my @db_dirs = get_daisy_db_dirs();

    foreach my $db_dir (@db_dirs) {
	my $tel_set_count = get_tel_set_count($db_dir);
	if ($tel_set_count > $edir_max_tel_sets) {
	    my $tel_sets_removed = eu_rm_old_tel_sets($db_dir, $edir_max_tel_sets);
	    loginfo("number of old tel sets purged: $tel_sets_removed");
	}
    }

    return(1);
}


#
# Count the number of tel sets in a given "backups" dir.
# Assumption: There is one "control.tel.prepatch.$timestamp" file
# per set of backed up tel files, so just count them to determine
# the number of backed up tel sets.
#
# Returns
#   number of sets
#
sub get_tel_set_count
{
    my ($db_dir) = @_;

    my @control_files = glob("$db_dir/backups/control.tel.prepatch.*");

    my $tel_set_count = scalar(@control_files);

    return($tel_set_count);
}


#
# Remove the oldest tel sets from specified dir keeping only the latest
# "n" sets where "n" is the number specified.
#
# Returns
#   number of sets removed
#
sub eu_rm_old_tel_sets
{
    my ($db_dir, $edir_max_tel_sets) = @_;

    # make a list with the "control.tel" file from each tel set and
    # verify there are some sets to remove
    my @control_files = glob("$db_dir/backups/control.tel.prepatch.*");
    if (scalar(@control_files) <= $edir_max_tel_sets) {
	return(0);
    }

    # make a list with the timestamp from each tel set and
    # verify there are some sets to remove
    my @timestamps = ();
    foreach my $filename (@control_files) {
        if ($filename =~ /.*\.tel\.prepatch\.(\d{14})/x) {
	    push(@timestamps, $1);
	}
    }
    my $tel_set_count = scalar(@timestamps);
    if ($tel_set_count <= $edir_max_tel_sets) {
	return(0);
    }

    # remove the excess tel sets
    my $remove_count = $tel_set_count - $edir_max_tel_sets;
    my @timestamps_sorted = sort(@timestamps);
    my @timestamps_to_remove = @timestamps_sorted[0 .. $remove_count-1];
    foreach (@timestamps_to_remove) {
	system("rm $db_dir/backups/*.$_");
    }

    return($remove_count);
}


#
# Run command with "system" function
#
# Returns
#   0 for success
#   -1 if timeout occurred
#   -2 some other unexpected event occurred
#   > 0 is exit status of command
#
sub eu_system_with_timeout
{
    my ($cmd, $timeout) = @_;

    my $rc = eval {
	local $SIG{ALRM} = sub { croak "alarm timeout" };
	alarm $timeout;
	my $cmd_rc = mysystem("$cmd");
	alarm 0;
	$cmd_rc;
    };

    #
    # if $@ is not empty, then either the timeout happened or
    # something else unusual.
    #
    if ($@) {
	# timeout
	if ($@ =~ /alarm timeout/) {
	    $rc = -1;
	    logerror("timeout after $timeout seconds while running $cmd: $@");
	}
	# something else
	else {
	    # for this case, the alarm might need to be cancelled
	    alarm 0;
	    $rc = -2;
	    logerror("unexpected event while running $cmd: $@");
	}
    }

    return($rc);
}



#
# Given a record name, return the record number.  Record numbers are
# numbered from 0 to ((nr of recs) - 1);
#
# Returns
#   0 to ((nr of recs) - 1) on success
#   -1 if not
#
sub eu_summary_logfile_rec_name_to_rec_nr
{
    my ($rec_name) = @_;

    my $max_rec_nr = $SummaryLogFileRecCount - 1;
    for (my $i=0; $i <= $max_rec_nr; $i++) {
	if ($rec_name eq $SummaryLogFileRecs[$i]{rec_name}) {
	    return($i);
	}
    }

    return(-1);
}


sub eu_summary_logfile_set_rec_value_by_rec_nr
{
    my ($rec_nr, $rec_value) = @_;

    my $max_rec_nr = $SummaryLogFileRecCount - 1;

    if ($rec_nr > $max_rec_nr) {
	logerror("rec number too large: $rec_nr (max: $max_rec_nr)\n");
    }
    else {
	$SummaryLogFileRecs[$rec_nr]{rec_value} = $rec_value;
    }

    return($rec_nr);
}


sub eu_summary_logfile_set_rec_value_by_rec_name
{
    my ($rec_name, $rec_value) = @_;

    my $rec_nr = eu_summary_logfile_rec_name_to_rec_nr($rec_name);
    if ($rec_nr >= 0) {
	eu_summary_logfile_set_rec_value_by_rec_nr($rec_nr, $rec_value);
    }
    else {
	logerror("unknown summary logfile record name: $rec_name");
    }

    return(1);
}


sub eu_summary_logfile_note_success
{
    eu_summary_logfile_set_rec_value_by_rec_name('Status', "SUCCESSFUL");

    return(1);
}


sub eu_summary_logfile_note_failure
{
    my ($failure_desc) = @_;

    my $failure_time = strftime("%a, %b %d, %Y at %H:%S", localtime());
    eu_summary_logfile_set_rec_value_by_rec_name('Status', "FAILURE");
    eu_summary_logfile_set_rec_value_by_rec_name('FailDate', $failure_time);
    eu_summary_logfile_set_rec_value_by_rec_name('FailDesc', $failure_desc);

    return(1);
}


sub eu_summary_logfile_generate
{
    my $rc = 1;
    my $edir_date = $EDIR_YEAR . $EDIR_BOOK;

    # report time of last delta request
    eu_summary_logfile_determine_last_delta_request();

    # report the last edir update
    eu_summary_logfile_determine_last_edir_update();

    # report the last delta directory and product sequence numbers.
    eu_summary_logfile_determine_delta_seq_nrs($edir_date, $EDIR_DELTADIR);

    if (open(my $slf, ">>", $SUMMARY_LOGFILE_DEF)) {

	# if there is already some contents, then output separator
	if (-s $SUMMARY_LOGFILE_DEF) {
	    print $slf "\n";
	}

	my $rec_count = eu_summary_logfile_put_records($slf);

	close($slf);

	loginfo("number of records written to summary log file: $rec_count");
    }
    else {
	$rc = 0;
    }

    return($rc);
}


sub eu_summary_logfile_put_records
{
    my ($slf) = @_;

    my $max_rec_nr = $SummaryLogFileRecCount - 1;
    my $success = ($SummaryLogFileRecs[0]{rec_value} eq "SUCCESSFUL");

    my $i = 0;
    while ($i <= $max_rec_nr) {
	my $rec_desc = $SummaryLogFileRecs[$i]{rec_desc};
	my $rec_value = $SummaryLogFileRecs[$i]{rec_value};

	# skip failure records when status is success
	if ($success) {
	    if ($rec_desc =~ /^FAILURE/x) {
		next;
	    }
	}
	print $slf "$rec_desc: $rec_value\n";

	$i++;
    }

    return($i);
}


sub eu_summary_logfile_determine_last_delta_request
{
    my $last_delta_req_time = "not available";

    if (-f $EDIR_DELTA_LAST_REQ) {
	my $mtime = 9;  # see docs for stat built-in
	my $file_mtime = (stat($EDIR_DELTA_LAST_REQ))[$mtime];
	$last_delta_req_time = strftime("%a, %b %d, %Y at %H:%S", localtime($file_mtime));
    }
    else {
	logerror("time of last delta request not available");
    }

    my $rec_nr = eu_summary_logfile_rec_name_to_rec_nr('DeltaReqDate');
    $SummaryLogFileRecs[$rec_nr]{rec_value} = $last_delta_req_time;

    return(1);
}


sub eu_summary_logfile_determine_last_edir_update
{
    my $edir_control_file = "/d/daisy/control.tel";

    my $mtime = 9;  # see docs for stat built-in
    my $file_mtime = (stat($edir_control_file))[$mtime];
    my $last_edir_update = strftime("%a, %b %d, %Y at %H:%S", localtime($file_mtime));

    my $rec_nr = eu_summary_logfile_rec_name_to_rec_nr('DirUpdate');
    $SummaryLogFileRecs[$rec_nr]{rec_value} = $last_edir_update;

    $rec_nr = eu_summary_logfile_rec_name_to_rec_nr('ProdUpdate');
    $SummaryLogFileRecs[$rec_nr]{rec_value} = $last_edir_update;

    return(1);
}


sub eu_summary_logfile_determine_delta_seq_nrs
{
    my ($edir_date, $delta_dir) = @_;

    my @delta_types_tab =(
	{delta_type => 'Directory', delta_rec_name => 'DirSeqNr'},
	{delta_type => 'Product',   delta_rec_name => 'ProdSeqNr'},
    );

    for my $i (0 .. $#delta_types_tab) {
	my $delta_type = $delta_types_tab[$i]{delta_type};
        my $last_seq_nr = eu_summary_logfile_get_delta_seq_nr($edir_date, $delta_dir, $delta_type);

	my $rec_name = $delta_types_tab[$i]{delta_rec_name};
	my $rec_nr = eu_summary_logfile_rec_name_to_rec_nr($rec_name);

	$SummaryLogFileRecs[$rec_nr]{rec_value} = $last_seq_nr;
    }

    return(1);
}

sub eu_summary_logfile_get_delta_seq_nr
{
    my ($edir_date, $delta_dir, $delta_type) = @_;

    my $delta_file_re = $delta_dir . '/' . $edir_date . "Delta" . $delta_type . "*.xml";
    my @delta_files = glob($delta_file_re);

    my $last_seq_nr = 0;
    foreach (@delta_files) {
	my $delta_file = basename($_);
	if ($delta_file =~ /^\d{4}\w{3}(Delta)($delta_type)\d{20}_(\d{20})\.xml/x) {
	    my $seq_nr = $3;
	    if ($seq_nr > $last_seq_nr) {
		$last_seq_nr = $seq_nr;
	    }
	}
    }

    # remove leading 0's
    $last_seq_nr =~ s/^0*(\d+)$/$1/x;

    return($last_seq_nr);
}


sub eu_select_exit_status_desc
{
    my ($exit_status) = @_;

    my %exit_status_desc_table = (
	1  => "SIGHUP Hangup",
	2  => "SIGINT Interrupt",
	3  => "SIGQUIT Quit",
	4  => "SIGILL Illegal instruction",
	5  => "SIGTRAP Trace trap",
	6  => "SIGABRT Abort",
	7  => "SIGBUS BUS error",
	8  => "SIGFPE Floating-point exception",
	9  => "SIGKILL Kill, unblockable",
	10 => "SIGUSR1 User-defined signal 1",
	11 => "SIGSEGV Segmentation violation",
	12 => "SIGUSR2 User-defined signal 2",
	13 => "SIGPIPE Broken pipe",
	14 => "SIGALRM Alarm clock",
	15 => "SIGTERM Termination",
	16 => "SIGTKFLT Stack fault",
	17 => "SIGCHLD Child status has changed",
	18 => "SIGCONT Continue",
	19 => "SIGSTOP Stop, unblockable",
	20 => "SIGTSTP Keyboard stop",
	21 => "SIGTTIN Background read from tty",
	22 => "SIGTTOU Background write to tty",
	23 => "SIGURG Urgent condition on socket",
	24 => "SIGXCPU CPU limit exceeded",
	25 => "SIGXFSZ File size limit exceeded",
	26 => "SIGVTALRM Virtual alarm clock",
	27 => "SIGPROF Profiling alarm clock",
	28 => "SIGWINCH Window size change",
	29 => "SIGIO I/O now possible",
	30 => "SIGPWR Power failure restart",
	31 => "SIGSYS Bad system call",
	101 => "exit status 101",
	102 => "exit status 102",
	103 => "exit status 103",
	104 => "exit status 104",
	105 => "exit status 105",
	106 => "exit status 106",
	107 => "exit status 107",
	108 => "exit status 108",
	109 => "exit status 109",
    );

    my $exit_status_desc = "unknown";

    if ($exit_status >= 0 && defined($exit_status_desc_table{$exit_status})) {
	$exit_status_desc = $exit_status_desc_table{$exit_status};
    }

    return($exit_status_desc);
}


#########################################
#########   F U N C T I O N S   #########
#########################################

sub usage
{
    print "Usage:\n";
    print "$PROGNAME --version\n";
    print "$PROGNAME --help\n";
    print "$PROGNAME [options]\n";
    print "where options are:\n";
    print "\t--verbose                    output extra info\n";
    print "\t--edir-config=s              path to config file\n";
    print "\t--edir-basedir=s             path to top of installation tree\n";
    print "\t--edir-book=s                name of book (jan, may, sep)\n";
    print "\t--edir-year=s                year of book (YYYY)\n";
    print "\t--edir-getdeltas-timeout=s   wait for GetDeltas this many seconds\n";
    print "\t--edir-maketel-timeout=s     wait for maketel this many seconds\n";
    print "\t--report-config              output config file lines as they are read\n";
    print "\t--force-maketel              run the maketel program\n";
    print "\t--lockfile-remove            remove the lockfile\n";
    print "\t--summary-logfile            generate a summary log file\n";
    print "\t--delay=n                    specify a delay before getting deltas\n";
    print "\n";

    return(1);
}


#
# Return a list of verified POS dirs.  For Daisy, remember, there can be
# more than one daisy database dir!
#
sub get_daisy_db_dirs
{
    my @daisy_dirs = ();
    my $daisy_top_dir = '/d';

    # Daisy POS can have one or more database dirs
    my @daisy_db_dirs = glob("$daisy_top_dir/*");
    for my $daisy_db_dir (@daisy_db_dirs) {

	# must be a directory
	next unless (-d $daisy_db_dir);

	# skip old daisy dirs
	next if ($daisy_db_dir =~ /.+-\d{12}$/x);

	# must contain the magic files
	next unless(-e "$daisy_db_dir/flordat.tel");
	next unless(-e "$daisy_db_dir/control.dsy");

	# must be daisy 8.0+
	next unless (-d "$daisy_db_dir/bin");

	push(@daisy_dirs, $daisy_db_dir);
    }

    return(@daisy_dirs);
}


#
# Return the percentage of disk being used.
# This gives us a clue as to whether the disk is already "at capacity".
#
# Returns
#   positive integer on success
#   -1 on error
#
sub get_diskused
{
    my ($directory) = @_;

    my $returnval = -1;

    if (-d "$directory") {
	if (open(my $pipe, '-|', "df -h $directory")) {
	    while (<$pipe>) {
		if (/\s+(\d+)\%\s+/x) {
		    $returnval = $1;
		    last;
		}
	    }
	    close($pipe);
	}
    }

    return($returnval);
}


sub loginfo
{
    my ($message) = @_;

    chomp($message);
    $message = "<I> $message";
    logit($message);

    return(1);
}


sub logerror
{
    my ($message) = @_;

    chomp($message);
    $message = "<E> $message";
    logit($message);

    return(1);
}


sub logit
{
    my ($message) = @_;

    my $logtime = strftime("%Y-%m-%d %H:%M:%S", localtime());

    chomp($message);
    if (open(my $logfh, '>>', $LOGFILE)) {
	print($logfh "$logtime ($PROGNAME-$PID) $message\n");
	close($logfh);
    }
    else {
	print "$logtime ($PROGNAME-$PID) $message\n";
    }

    if ($VERBOSE) {
	print $message . "\n";
    }

    return(1);
}


#
# Subroutine to run system commands.  This logs an entry.
#
sub mysystem
{
    my ($command) = @_;

    my $rc = $EMPTY;

    system("$command >> $LOGFILE 2>> $LOGFILE");
    if (WIFEXITED($?)) {
	$rc = WEXITSTATUS($?);
	loginfo("child terminated normally with status: $rc");
    }
    elsif (WIFSIGNALED($?)) {
	$rc = WTERMSIG($?);
	loginfo("child terminated because of signal: $rc");
    }
    else {
	$rc = $?;
    }

    loginfo("$command Returned: $rc");

    return ($rc)
}


#################################################
#########   D O C U M E N T A T I O N   #########
#################################################

__END__

=pod

=head2 NAME

I<edir_update.pl> - script to update the Daisy POS edirectory files


=head2 VERSION

This documentation refers to version: $Revision: 1.14 $


=head2 OPTIONS

=over 4

=item B<--version>

Output the version number of the script and exit.

=item B<--help>

Output a short help message and exit.

=item B<--verbose>

Modifier: report extra information.

=item B<--edir-config>

Specify path to F<edir_update.pl> config file.

=item B<--edir-basedir>

Specify an alternative top level directory for eDirectory files.

=item B<--edir-book>

Specify the "book" of the eDirectory.

=item B<--edir-year>

Specify the "year" of the eDirectory.

=item B<--edir-getdeltas-timeout>

Specify a timeout for the F<GetDeltas> command.

=item B<--edir-maketel-timeout>

Specify a timeout for the F<maketel> command.

=item B<--force-maketel>

Force execution of the F<maketel> program without regard to time.

=item B<--lockfile-remove>

Clear out an errant lockfile.

=item B<--delay=n>

Allows for a specific delay or no deley before calling F<GetDeltas>

=item B<--savedtelsets=n>

Provides the specification for the number of sets of "tel files"
to be kept in the backup directory.

=back


=head2 DESCRIPTION

Each Daisy POS has within it a set of eDirectory files.
These files consist of florist directory data files,
floral product data files, programs, and config files.
There are 3 comprehensive eDirectory distributions per year
to each Daisy POS.

An eDirectory distribution is designated with a book name and a year.
The book names are "jan", "may", and "sep" and
the "year" merely consists of the 4 digit form of the year.
Thus, for example, a Daisy POS might have the "2013jan" eDirectory installed.
In May, it would be updated to the "2013may" eDirectory.
The F<edir_installbase.pl> script is used to install one of these
eDirectory distributions.

With eDirectory distributions installed on a Daisy system only every
3 months, the data can become stale quickly.
In order to keep the data in an eDirectory distribution up to date
with a much finer granularity, the F<edir_udpate.pl> script is used
to update the eDirectory on a daily basis.

=head2 Overview of an eDir Update

The F<edir_update.pl> script is called via a cron job every 4 hours,
at the hours of 2am, 6am, 10am, 2pm, 6pm and 10pm.
As the first step in the upgrade process,
it runs the F<GetDeltas> program to download from the Teleflora "Delta Force"
web site any changes to the currently installed base set of
florist directory and floral product files.
These downloads are "XML" files and are called the "delta files";
these delta files are accumulated and retained to be used in the
next step of the workflow.
As mentioned above, this download step is repeated every 4 hours.
On any given download, there may be zero or more delta files downloaded.
Usually, in a 24 hour period, there will be several "delta" files
accumulated.

Once a day, when the 2am instance of the cron jobs runs,
F<edir_update.pl> script runs the F<maketel> program.
When the F<maketel> program runs, it combines the base set of
eDirectory files with the delta files accumulated in the last
24 hours period, to produce a new set of eDirectory files.
After a successful F<maketel> run,
the new set of eDirectory files are installed in the standard
location for a Daisy POS using the standard eDirectory "applypatch" methodology.

=head2 Details of an eDir Update

Here is a detailed outline of what F<edir_update.pl> does.

=over 4

=item 0.

The script parses the command line and sets all the values
for any command line options present.
If there are any command line arguments,
the script outputs a usage message and exits C<$EXIT_COMMAND_LINE>.
The script writes the name of the program and the version string
for the program to the log file.
The "book" and the "year" info is written to the summary log database.

=item 1.

The script verifies that the Daisy file system, F</d>, is not
100% full.
If it is, then a message is written to the log file and
the summary log database and the script exits with an
exit status of C<$EXIT_DISK_SPACE>.

=item 2.

The script reads the config file, either the default config
file at path F</d/daisy/config/edir_update.conf> or
the one specified with the I<--edir-conf> command line option.
Any command line option values that were specified that are in common
with lines in the config file are overwritten by the values in the
config file.

=item 3.

The script verifies that either the default base directory
for the eDirectory files, F</d/edirectories>, or
the directory specified via the I<--edir-basedir> command line option exists.
If not, a message is written to the log file and
the summary log database and the script exits with an
exit status of C<$EXIT_EDIR_TREE>.

=item 4.

If the I<--lockfile-remove> is specified, the lockfile is removed.

=item 5.

Signals are set to ignore and
the script verifies that the current instance of the script is
the only one running by testing the existence of the lock file
F</d/daisy/config/edir_update.lock>.
If the lockfile exists, then signals are reset and
a message is written to the log file and
the summary log database and the script exits with an
exit status of C<$EXIT_LOCKFILE_MAKE_ERR>.

=item 6.

The script writes the process id to the file
F</d/daisy/config/edir_update.pid>.

=item 7.

The heart of the matter is taken up finally.
The F<edir_update()> function is called to actually do the
update of the eDirectory files.
If the function returns with a value of C<$EXIT_OK>,
success messages are written to the log file and
the summary log database.
Otherwise, an error message is written to the log file and
the summary log database.

=item 8.

The script cleans up by removing old sets of "tel files"
in the "backups" directory of each existing Daisy database dirs;
only the last 3 sets of "tel files" are preserved.
Also, the script removes the pid file, the lock file, and
resets signals.
It then writes the exit status to the log file.

=item 9.

The summary log database is written to the summary log file.

=item 10.

The script exits with the exit status returned by F<edir_udpate()>.

=back

=head2 Command Line Options

The B<--edir-config> command line option allows the specification
of a path to a config file for the F<edir_update.pl> script.
Currently, the script does not support this file.

The B<--edir-basedir> command line option allows the specification
of an alternative top level directory for eDirectory files.
The default locations is F</d/edirectories>.

The B<--edir-book> command line option allows the "book" of the
eDirectory to be specified.
This command line option is usually always specified but does have
a default value of "jan".

The B<--edir-year> command line option allows the "year" of the
eDirectory to be specified.
This command line option is usually always specified but does have
a default value of "2013".

The B<--edir-getdeltas-timeout> command line option allows the
specification of a timeout for the F<GetDeltas> command.
The unit is in seconds with a default value of 600 seconds,
ie. 10 minutes.

The B<--edir-maketel-timeout> command line option allows the
specification of a timeout for the F<maketel> command.
The unit is in seconds with a default value of 600 seconds,
ie. 10 minutes.

The B<--force-maketel> command line option provides a way to
force the execution of the F<maketel> program without regard
to the time.
By default, the F<maketel> program is only executed if the
time when execution is to begin is between 2am and 3am.

The B<--lockfile-remove> command line option provides a way
to clear out an errant lockfile that may have been incorrectly
left behind and thus preventing execution of a new instance
of the F<edir_update.pl> script.

The B<--delay=n> command line option which allows a user
allows a user to specify either no delay or a specific delay
(in seconds) before the script calls the "GetDeltas" program.
This ability greatly helps out with testing.

Allows for a specific delay or no deley before calling F<GetDeltas>

The B<--savedtelsets=n> command line option provides the
specification for the number of sets of "tel files"
to be kept in the "backups" directory.
The default number of sets kept is 3.

=head2 Summary Logfile Format

Each run of the F<edir_update.pl> script concatenates a
set of lines to the summary log file which provide a quick
summary of the success or lack thereof of the run.
The summary log file is named F<Delta_Stats.log> and
is located in the standard log file directory for Daisy POS systems
which is normally F</d/daisy/log>.

All dates in the summary log file are of the format:

 <dayname>, <monthname> <monthday>, <year> at <hour>:<min>

where

=over 4

=item <dayname>

    Mon | Tue | Wed | Thu | Fri | Sat | Sun

=item <monthname>

    Jan | Feb | Mar | Apr | May | Jun | Jul | Aug | Sep | Oct | Nov | Dec

=item <monthday

    1 .. 31

=item <year>

    2013 .. 2038

=item <hour>

    01 .. 23
    
=item <min>

    01 .. 59

=back
 
All lines are in the format of text - a description -
followed by COLON and a SPACE, and then a value
which is either text, a date, or a number.

=over 4

=item line 1

    text of "CURRENT UPDATE STATUS" with a value of "SUCCESSFUL" | "FAILURE".

=item line 2

    text of "CURRENT BOOK NAME" with a value of "JAN" | "MAY" | "SEP".

=item line 3

    text of "CURRENT BOOK YEAR" with a value of <yyyy>.

=item line 4

    text of "LAST DELTA REQUEST DATE" with a value of <date>.

=item line 5

    text of "LAST SUCCESSFUL DIR UPDATE" with a value of <date>.

=item line 6

    text of "LAST SUCCESSFUL PRODUCT UPDATE" with a value of <date>.

=item line 7

    text of "LAST SUCCESSFUL DIR SEQ NR" with a value of <nnnnnnnnnnnnnnnn>.

=item line 8

    text of "LAST SUCCESSFUL PROD SEQ NR" with a value of <nnnnnnnnnnnnnnnn>.

=back


If value of line 1 is "FAILURE", then there will be several more lines.

=over 4

=item line 9

    text of "FAILURE DATE" with a value of <date>

=item line 10 

    text of "FAILURE DESCRIPTION" with a value of <text>

=item line 11 

    text of "FAILURE EXIT STATUS GETDELTAS" with a value of <text>

=item line 12 

    text of "FAILURE EXIT STATUS MAKETEL" with a value of <text>

=back


=head2 EXAMPLES

The following command will run F<edir_update.pl> -
the user should be the "tfsupport" account or
any Daisy admin account.
Other than log files, all files required to exist, or downloaded, or
generated will be located in the directory F</d/edirectories/2012ndj>.

 $ perl edir_update.pl --edir-book=ndj --edir-year=2012


=head2 FILES

=over 4

=item F</d/daisy>

The default Daisy database directory - this directory must exist and
contain a standard Daisy POS installation.

=item F</d/daisy/backups>

The "backups" directory is the location of previous sets of "tel files".

=item F</d/edirectories>

This directory is the the default top level directory for all eDirectory files.
Contained in this directory will be a directory for each eDirectory release.
This location may be changed with the B<edir-basedir> command line option.

=item F</d/edirectories/$year$book>

This directory contains all the files concerning a particular eDirectory.
The symbol F<$year> stands for the year of the eDirectory and
the symbol F<$date> stands for the name of the eDirectory.
An eDirectory is specified by it's "year" and it's "book" name.
For example, the files for the eDirectory for year "2013" and
book "may" will be located in the directory F</d/edirectories/2013may>.

=item F</d/edirectories/$year$book/config>

This directory contains all the config files for the eDirectory
that corresponds to the "year" and "date" specified.

=item F<edir_update.lock>

This file is located in the F<config> directory for a particular eDirectory.
The lockfile is used to prevent more than one instance of this script from
running simultaneously.

=item F<edir_update.pid>

This file is located in the F<config> directory for a particular eDirectory.
When this script is executed, the process id of the script is written to this file.

=item F<edir_update_delta_count.txt>

This file is located in the F<config> directory for a particular eDirectory.
After each successful run of the F<GetDeltas> program,
the number of delta files are counted and the count written to this file.
This number is used to decide if any new delta files were downloaded by the
F<GetDeltas> program.

=item F<edir_update_delta_last_request.txt>

This file is located in the F<config> directory for a particular eDirectory.
The date of the last successful run of the F<GetDeltas> program is recorded
as the mtime of this file.

=item F<edir_update.conf>

This file is located in the F<config> directory for a particular eDirectory.
Currently, this file is not used, but may be used in the future to
contain config items that could influence the execution of this script.

=item F<GetDeltas.conf>

This file is located in the F<config> directory for a particular eDirectory.
Currently, options for the F<GetDeltas> program can only be communicated to it
via a config file.
This config file and it's content are generated before execing F<GetDeltas> and
then specified to the F<GetDeltas> program via it's F<--config> command line option.

=item F<maketel.conf>

This file is located in the F<config> directory for a particular eDirectory.
Currently, this config file is not used since the F<maketel> program observes
command line options which make use of a config file unnecessary.

=item F</d/daisy/log/edir_update.log>

The general log file produced by this script.
The log file will contain log messages from this script,
any output from the F<GetDeltas> program, and
any output from the F<maketel> program.

=item F</d/daisy/log/Delta_Stats.log>

The summary log file produced by this script.
The contents of this log file is specified in the
B<DESCRIPTION> section above.


=back


=head2 DIAGNOSTICS

=over 4

=item Exit status 0 ($EXIT_OK)

Successful completion.

=item Exit status 1 ($EXIT_COMMAND_LINE)

In general, there was an issue with the syntax of the command line.

=item Exit status 3 ($EXIT_DISK_SPACE)

There is not enough disk space to install edir archive.

=item Exit status 4 ($EXIT_EDIR_TREE)

The edir directory does not exist.

=item Exit status 5 ($EXIT_SUMMARY_LOGFILE)

The summary logfile could not be opened.

=item Exit status 6 ($EXIT_CMD_MISSING)

The edir command file does not exist.

=item Exit status 7 ($EXIT_CMD_PERMS)

The edir command file does not have the correct file perms.

=item Exit status 8 ($EXIT_GETDELTAS_TIMEOUT)

The F<GetDeltas> command execution ended from time out.

=item Exit status 9 ($EXIT_GETDELTAS_EXCEPTION)

The F<GetDeltas> command execution ended from an unexpected exception.

=item Exit status 10 ($EXIT_GETDELTAS_ERROR)

The F<GetDeltas> command exited with non-zero exit status.

=item Exit status 11 ($EXIT_MAKETEL_TIMEOUT)

The F<maketel> command execution ended from time out.

=item Exit status 12 ($EXIT_MAKETEL_EXCEPTION)

The F<maketel> command execution ended from an unexpected exception.

=item Exit status 13 ($EXIT_MAKETEL_ERROR)

The F<GetDeltas> command exited with non-zero exit status.

=item Exit status 14 ($EXIT_NO_NEW_DELTAS)

The F<GetDeltas> command ran but there were no new delta files downloaded
so there was no reason to make new "tel" files.

=item Exit status 15 ($EXIT_PATCH_DIR)

After the deltas are downloaded and the "tel" files are generated,
the script must make a "patch" file which will be used with
F<applypatch.pl> to install the new "tel" files.
The script attempts to make a working directory for holding the
"tel" files that will go into the patch.
If there is an error making this directory,
the script exits with exit status $EXIT_PATCH_DIR.

=item Exit status 16 ($EXIT_PATCH_BINDIR)

Referring to the explanation for $EXIT_PATCH_DIR,
the script attempts to make a directory to hold the binary files
that will go into the patch file, similar to the directory that holds the
"tel" files.
If there is an error making this directory,
the script exits with exit status $EXIT_PATCH_BINDIR.

=item Exit status 17 ($EXIT_MAKE_PATCH)

The F<makepatch.pl> command failed.
The log will report the exact command string and
the returned exit status.

=item Exit status 18 ($EXIT_PATCH_NOT_FOUND)

=item Exit status 19 ($EXIT_APPLY_PATCH)

=item Exit status 20 ($EXIT_LOCKFILE_MAKE_ERR)

An attempt to atomically verify that there was no existing lockfile and
to make a new lockfile failed... the assumption is that there is an
existing instance of the script already running.

=item Exit status 21 ($EXIT_LOCKFILE_RM_ERR)

The lockfile exists but could not be removed as requested via the
command line option.

=item Exit status 22 ($EXIT_GETDELTAS_CONF)

In order to execute the F<GetDeltas> program with non-default options,
a config file must be generated, the options written to the config file, and
the config file path specified on the F<GetDeltas> command line.
If the config file can not be generated,
the script exits with exit status $EXIT_GETDELTAS_CONF.

=back


=head2 SEE ALSO

I<GetDeltas(1)>, I<maketel(1)>, I<edir_installbase.pl>, I<applypatch.pl>


=cut
