#!/usr/bin/perl
#
# $Revision: 1.2 $
# Copyright 2010 Teleflora
# 
# This script intended to automate the process of creating code patches, specifically
# addressing PCI 7.1
#
# The patch naming convention is as follows:
#
#	the name starts with "DAISY-"
#	the --build option provides the quarter name, and year
#		the "FD" suffix for the value of the --build option
#		stands for "floral directory"
#	a --name option of "TEL" means just the Teleflora files are
#		contained in the patch
#	a --name option of "ALL" means both Teleflora and FTD types
#		of files
#
# Example invocation for the Daisy edir update patch for FMA 2011 which
# contains both the Teleflora files and the FTD files:
#
# makepatch.pl --build "FMA11FD" --name ALL  \
#	categories.tel citydat.tel cityidx.tel control.tel extproducts.tel \
#	fac_dat.tel fac_idx.tel flordat.tel floridx.tel occasions.tel \
#	products.tel variants.tel zip_dat.tel zip_idx.tel \
#	bin/postpatch.pl
#
# The resulting patch file would be named:
#
#	DAISY-FMA11FD_ALL.patch
#

use strict;
use warnings;
use Getopt::Long;
use POSIX;
use File::Basename;

my $CVS_REVISION = '$Revision: 1.2 $';
my $PROGNAME = basename($0);

#
# Command line options.
#
my $VERSION = 0;
my $HELP = 0;
my $LIST = 0;
my $BUILD = "";
my $NAME = "";
my $SHOP = "";
my $APPLYPKG = 0;
my $ENCRYPT = 0;
my $DAISY = 0;

#
# Global variables.
#
my @PATCHFILES = ();
my $PATCH_FILENAME = "";


GetOptions(
	"version" => \$VERSION,
	"help" => \$HELP,
	"build:s" => \$BUILD,
	"name:s" => \$NAME,
	"shop:s" => \$SHOP,
	"list" => \$LIST,
	"show" => \$LIST,
	"applypkg" => \$APPLYPKG,
	"encrypt" => \$ENCRYPT,
	"daisy" => \$DAISY,
) || die "Error: invalid command line option, exiting...\n";

# --version
if($VERSION != 0) {
	print("$CVS_REVISION\n");
	exit(0);
}

# --help
if($HELP != 0) {
	usage();
	exit(0);
}

if($LIST != 0) {
	list_patch($ARGV[0]);
	exit(0);
}

if($APPLYPKG != 0) {
	create_applypatch_tar();
	exit(0);
}


# --build
if($BUILD eq "") {
	usage();
	print("Error: Must specify a build name.\n");
	exit(1);
}


# --name
if($NAME eq "") {
	usage();
	print("Error: Must specify a patch name.\n");
	exit(1);
}


if($#ARGV < 0) {
	usage();
	print("Error: Must specify at least one source file/directory to go within the patch.\n");
	exit(2);
}


# Do all of the specified files actually exist?
@PATCHFILES = verify_patch_files(\@ARGV);


$PATCH_FILENAME = form_patch_filename($BUILD, $NAME, $SHOP);
if ($PATCH_FILENAME eq "") {
	print("Error: can't form patch filename\n");
	exit(1);
}

# Does our patch file name already exist?
if (-f $PATCH_FILENAME) {
	usage();
	print("Error: patch file already exists: $PATCH_FILENAME\n");
	exit(1);
}

make_patch_file(\@PATCHFILES, $PATCH_FILENAME);


exit(0);

#########################################################
#########################################################
#########################################################


sub usage
{
	print("Usage:\n");
	print("Make a Daisy \".patch\" file.\n");
	print("$PROGNAME --help\n");
	print("$PROGNAME --version\n");
	print("$PROGNAME --applypkg # Make a 'DAISY-applypatch.tar' file for Altiris.\n");
	print("$PROGNAME --list DAISY-build_tracker_shop.patch # List contents of patch.\n");
	print("$PROGNAME --show DAISY-build_tracker_shop.patch # List contents of patch.\n");
	print("\n");
	print("$PROGNAME --daisy --build \"8.1.3\" --name \"A\" bin/dsyperms.pl bin/dsyuser.pl ...\n");
	print("\n");
	print("\n");
}





#
# Verify that the files to go into this patch are valid.
# This is also where we could do some path 'translation'; for instance,
# some files which live in 'shellScripts' under CVS ultimately live in 'bin' 
# on a customer's box.
#
sub verify_patch_files
{
	my @argvlist = @{$_[0]};
	my @patchfiles = ();

	foreach my $thisfile (@argvlist) {

		if ( (! -f $thisfile) && (! -d $thisfile) ) {
			usage();
			print("\n\n");
			print("Error: not a file or directory: $thisfile\n");
			exit(3);
		}
		push(@patchfiles, $thisfile);
	}

	return(@patchfiles);
}


sub list_patch
{
	my $patchfile = $_[0];

	system("gzip -dc $patchfile | tar -tvf - ");

}


#
# For instructions in re-creating OpenSSL keys See also:
# http://www.devco.net/archives/2006/02/13/public_-_private_key_encryption_using_openssl.php
#
# The process is basically....
#
# rtidev$ openssl genrsa -out private.pem 1024
# rtidev$ openssl rsa -in private.pem -out public.pem -outform PEM -pubout
#
# Take the "private" key and place it inside of the "applypatch.pl" script.
# Take the "public" key and place it inside of the "makepatch.pl" script.
#
sub make_pubkey_file
{
	my $public_key_file = "";

	$public_key_file = qx(mktemp /tmp/makepatch.XXXXXXX);
	chomp($public_key_file);
	if ($public_key_file ne "") {

		open(KEYFILE, "> $public_key_file");
		print(KEYFILE << 'EOF');
-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCm22FT+P12E2ZyfvhtnSAkDSBs
1c1Bj+eqdZS1eVrWVHTEyWr9zw17fpXalfy6RH7e13KInR3xpVbcQa8j5Wu6YFlG
cohXjVtQnrgvxEZYJiXI465iSO8/W3tAZnMoHoAkSYHEvf5eON68zV1NIBjAWYuj
KqzBBhJwAfhNpZ4+rwIDAQAB
-----END PUBLIC KEY-----
EOF
		close(KEYFILE);
	}

	return($public_key_file);
}


sub form_patch_filename
{
	my $build = $_[0];
	my $name = $_[1];
	my $shop = $_[2];
	my $patch_filename = "";

	#
	# Package names will be...
	# DAISY-build_name.patch
	#
	if ($DAISY) {
		$patch_filename = 'DAISY-';
	}

	$patch_filename .= $build . '_' . $name;

	if ($shop ne "") {
		$patch_filename .= '_' . $shop;
	}

	$patch_filename .= '.patch';

	return($patch_filename);
}


sub make_patch_file
{
	my @patchfiles = @{$_[0]};
	my $patch_filename = $_[1];
	my $pubkey_file = "";

	if ($ENCRYPT) {
		$pubkey_file = make_pubkey_file();
		if ($pubkey_file eq "") {
			print("Error: can't make public key file\n");
			$ENCRYPT = 0;
		}
	}

	#
	# Form the tar command - the encryption step is optional and
	# can be specified with a command line option.
	#
	my $tar_cmd = "tar -cf - @patchfiles";
	my $encrypt_cmd = "openssl rsautl -encrypt -inkey $pubkey_file -pubin";
	my $compress_cmd = "gzip -9";

	if ($ENCRYPT) {
		$tar_cmd .= " | " . $encrypt_cmd;
	}
	$tar_cmd .= " | " . $compress_cmd;

	system("$tar_cmd > $patch_filename");

	if (-s $patch_filename) {
		print("New patch file: $patch_filename\n");
	}

	if ($ENCRYPT) {
		unlink($pubkey_file);
	}
}


#
# Create tarball which Altiris uses to apply patches.
# This should be a rare case; hopefully.
#
sub create_applypatch_tar
{
	my $applypkg = "";

	# Make sure "applypatch.pl" script is here.
	if(! -f "applypatch.pl") {
		usage();
		print("Error: Could not find \"./applypatch.pl\" script in current directory.\n");
		exit(5);
	}

	# Create our Apply Patch "package"
	$applypkg  = "DAISY-applypatch.tar";
	system("tar -C \"bin\" -cf $applypkg applypatch.pl");
	if(-s "$applypkg") {
		print("Created Applypatch Package: $applypkg\n");
	}

	return(0);
}

