#!/usr/bin/perl
# 
# $Revision: 1.40 $
# Copyright 2011 Teleflora.
#
# The script constitutes the postpatch script for the process of
# applying patches to Daisy systems and for converting a Daisy
# system from the Global East payment system to Elavon.
#
# Called from applypatch.pl with arguments of:
#
#	ARGV[0] = path to current daisy db dir
#	ARGV[1] = time stamp
#	ARGV[2] = list of patch files
#

use strict;
use warnings;
use Getopt::Long;
use POSIX;
use English;
use File::Basename;
use File::Spec;

our $VERSION = 1.15;
my $CVS_REVISION = '$Revision: 1.40 $';
my $PROGNAME = basename($0);

#
# Command line options
#
my $DEF_EDIRUPDATE        = 1;
my $DEF_INSTALL_FTD_FILES = 0;
my $DEF_DEBUG             = 0;

my $CLO_HELP              = 0;
my $CLO_VERSION           = 0;
my $CLO_EDIRUPDATE        = $DEF_EDIRUPDATE;
my $CLO_INSTALL_FTD_FILES = $DEF_INSTALL_FTD_FILES;
my $CLO_DEBUG             = $DEF_DEBUG;

#
# Constants
#
my $EMPTY_STR   = q{};
my $SPACE       = q{ };
my $COMMA       = q{,};
my $COLON       = q{:};
my $ATSIGN      = q{@};
my $DOT         = q{.};
my $DOTDOT      = q{..};
my $SLASH       = q{/};
my $DASH        = q{-};
my $EQUALS      = q{=};
my $SINGLEQUOTE = q{'};

#
# Global variables
#
my @SUPPORTED_PLATFORMS = qw(
	RHEL5
	RHEL6
	RHEL7
    );

#
# Daisy file locations
#
my $DAISY_TOPDIR          = '/d';
my $DAISY_DIR             = File::Spec->catdir($DAISY_TOPDIR,  'daisy');
my $DAISY_BINDIR          = File::Spec->catdir($DAISY_DIR, 'bin');
my $DAISY_CONFIGDIR       = File::Spec->catdir($DAISY_DIR, 'config');
my $DAISY_LOGDIR          = File::Spec->catdir($DAISY_DIR, 'log');

my $DAISY_BACKUP_DIRNAME  = 'backups';

my $FLORIST_DIR_NAME      = 'edirectories';
my $DAISY_FLORIST_DIR     = File::Spec->catdir($DAISY_TOPDIR, $FLORIST_DIR_NAME);

my $DOVE_CONTROL_FILENAME = 'dovectrl.pos';
my $DAISY_SHOPCODE_FILE   = File::Spec->catdir($DAISY_DIR, $DOVE_CONTROL_FILENAME);

my $PATCH_CONF_FILENAME   = 'applypatch.config';
my $PATCH_CONF_FILE       = File::Spec->catdir($DAISY_CONFIGDIR, $PATCH_CONF_FILENAME);

my $LOGFILE_NAME          = 'postpatch.log';
my $LOGFILE               = File::Spec->catdir($DAISY_LOGDIR, $LOGFILE_NAME);

my $FTD_SUBS_FILENAME     = 'FTD_Subscribers.txt';
my $FTD_FILES_PREFIX      = 'FTD_Files_';
my $FTD_FILES_SUFFIX      = '.tar';

my $DAISY_TOOLS_DIR       = File::Spec->catdir($DAISY_TOPDIR, 'ostools');
my $DAISY_TOOLS_BINDIR    = File::Spec->catdir($DAISY_TOOLS_DIR, 'bin');
my $DAISY_TOOLS_CONFDIR   = File::Spec->catdir($DAISY_TOOLS_DIR, 'config');

my @CONFIG_QA_SHOPCODE = ();

#
# Exit status values
#
my $EXIT_OK = 0;
my $EXIT_COMMAND_LINE = 1;
my $EXIT_MUST_BE_ROOT = 2;
my $EXIT_PATCH_VALIDATION = 3;
my $EXIT_POSTPATCH_CLEANUP = 8;
my $EXIT_STOP_DAISY = 9;
my $EXIT_START_DAISY = 10;
my $EXIT_PATCH_FAILED = 11;
my $EXIT_POSTPATCH = 12;
my $EXIT_EDIR_UPDATE = 14;
my $EXIT_BINARY_UPDATE = 15;


GetOptions(
	"help" => \$CLO_HELP,
	"version" => \$CLO_VERSION,
	"edirupdate!" => \$CLO_EDIRUPDATE,
	"install-ftd-files" => \$CLO_INSTALL_FTD_FILES,
	"debug" => \$CLO_DEBUG,
) || die "Error: invalid command line option, exiting...\n";



# --version
if ($CLO_VERSION) {
	print("$PROGNAME $CVS_REVISION\n");
	exit($EXIT_OK);
}

# --help
if ($CLO_HELP) {
	print("$PROGNAME $CVS_REVISION\n");
	pp_usage();
	exit($EXIT_OK);
}


my ($PATCHED_DAISY_DIR, $TIMESTAMP, @PATCH_FILES) = @ARGV;

if (defined($PATCHED_DAISY_DIR)) {
    if ($PATCHED_DAISY_DIR) {
	loginfo("[postpatch] path to patched Daisy dir: $PATCHED_DAISY_DIR");
    }
    else {
	logerror("[postpatch] first arg empty but is required to specify path of patched daisy dir");
	exit($EXIT_COMMAND_LINE);
    }
}
else {
    logerror("[postpatch] first arg not defined but is required to specify path of patched daisy dir");
    exit($EXIT_COMMAND_LINE);
}

if (defined($TIMESTAMP)) {
    if ($TIMESTAMP) {
	loginfo("[postpatch] time stamp of patched Daisy dir: $TIMESTAMP");
    }
    else {
	logerror("[postpatch] second arg is empty but is required to specify time stamp");
	exit($EXIT_COMMAND_LINE);
    }
}
else {
    logerror("[postpatch] second arg is not defined but is required to specify time stamp");
    exit($EXIT_COMMAND_LINE);
}

my $patch_files_count = scalar(@PATCH_FILES);
loginfo("[postpatch] number of patch files: $patch_files_count");

#
# Read the config file
#
pp_read_configfile(dsy_pathto_patch_config_file());

#
# perform post patch processing
#
exit(pp_main($PATCHED_DAISY_DIR));


sub pp_main
{
    my ($patched_daisy_dir) = @_;

    my $exit_status = $EXIT_OK;

    loginfo("----------------------------------------------------------------");
    loginfo("Begin $PROGNAME");
    loginfo("----------------------------------------------------------------");

    if ($CLO_EDIRUPDATE) {
	if (pp_postpatch_edir_update($patched_daisy_dir)) {
	    loginfo("[main] postpatch of florist directory update successful");
	    pp_remove_ftd_subs_file($patched_daisy_dir);
	}
	else {
	    logerror("[main] could not postpatch florist directory update");
	    $exit_status = $EXIT_EDIR_UPDATE;
	}
    }

    #
    # If there were new or updated binaries, remove the old ones, and
    # install the new ones.
    #
    if (pp_resolve_binaries($patched_daisy_dir)) {
	loginfo("[main] postpatch of new binaries successful");
    }
    else {
	logerror("[main] could not postpatch new binaries");
	$exit_status = $EXIT_BINARY_UPDATE;
    }


    #
    # Remove the postpatch.pl script we are currently running.
    #
    pp_remove_postpatch_script($patched_daisy_dir);

    loginfo("----------------------------------------------------------------");
    loginfo("Completed $PROGNAME");
    loginfo("----------------------------------------------------------------");

    if ($exit_status == $EXIT_OK) {
	my $daisy_dir = dsy_pathto_daisydir();
	chdir $daisy_dir;
	my $cmd = "$daisy_dir/logevent";
	my $msg = "florist directory in $patched_daisy_dir has been updated";
	if (-e $cmd) {
	    system("$cmd nEventSystemUpdate 0 \"\" SYS postpatch.pl \"$msg\"");
	}
    }

    return($exit_status);
}


#
# Daisy functions
#

sub dsy_pathto_topdir
{
    return($DAISY_TOPDIR);
}

sub dsy_pathto_daisydir
{
    return($DAISY_DIR);
}

sub dsy_pathto_configdir
{
    return($DAISY_CONFIGDIR);
}

sub dsy_pathto_logdir
{
    return($DAISY_LOGDIR);
}

sub dsy_pathto_shopcode_file
{
    return($DAISY_SHOPCODE_FILE);
}

sub dsy_pathto_patch_config_file
{
    return($PATCH_CONF_FILE);
}

sub dsy_pathto_ftd_subs_file
{
    my ($patched_daisy_dir) = @_;

    return(File::Spec->catdir($patched_daisy_dir, $FTD_SUBS_FILENAME));
}

sub dsy_pathto_florist_dir
{
    return($DAISY_FLORIST_DIR);
}

sub dsy_pathto_ftd_files_archive
{
    my $florist_dir = dsy_pathto_florist_dir();
    my $ftd_files_archive = dsy_nameof_ftd_files_archive();

    return(File::Spec->catdir($florist_dir, $ftd_files_archive));
}

sub dsy_nameof_backup_dir
{
    return($DAISY_BACKUP_DIRNAME);
}

sub dsy_nameof_ftd_files_archive
{
    my $ftd_files_archive  = $FTD_FILES_PREFIX . $TIMESTAMP . $FTD_FILES_SUFFIX;

    return($ftd_files_archive);
}

#######
#### Functions
#######


sub pp_usage
{
    print("Usage:\n");
    print("$PROGNAME --version\n");
    print("$PROGNAME --help\n");
    print("$PROGNAME [options] daisy_dir time_stamp list_of_files\n");
    print("\n");

    print("Options:\n");
    print("--[no]edirupdate             # install florist directory updates\n");
    print("--install-ftd-files          # install FTD files unconditionally\n");
    print("--debug                      # include debug output\n");
    print("\n");

    print("Defaults:\n");
    print("Install florist directory updates: " . (($DEF_EDIRUPDATE) ? "Yes" : "No") . "\n");
    print("Install FTD files: " . (($DEF_INSTALL_FTD_FILES) ? "Yes" : "No") . "\n");
    print("Include debug output: " . (($DEF_DEBUG) ? "Yes" : "No") . "\n");

    return(1);
}


#
# Read the config file located in default daisy db dir: /d/daisy/config
#
# Returns
#   1 on success
#   0 on error
#
sub pp_read_configfile
{
    my ($config_file) = @_;

    my $ml = '[read_configfile]';

    if ($config_file eq $EMPTY_STR) {
	return(1);
    }

    if (! -e $config_file) {
	loginfo("$ml config file does not exist: $config_file");
	return(0);
    }

    if (open(my $fh, '<', $config_file)) {
	while(<$fh>) {

	    # qa-shopcode=nnnnnn00
	    # Multiple instances of this line are allowed.
	    if (/^\s*qa-shopcode\s*=\s*([[:print:]]+)$/i) {
		    push(@CONFIG_QA_SHOPCODE, $1);
	    }
	}
	close($fh) or warn "$ml could not close config file: $config_file ($OS_ERROR)\n";
    }
    else {
	logerror("$ml could not open config file: $config_file");
	return(0);
    }

    return(1);
}


sub pp_remove_postpatch_script
{
    my ($patched_daisy_dir) = @_;

    my $postpatch_script = File::Spec->catdir($patched_daisy_dir, '/bin/postpatch.pl');
    my $rc = 1;
    my $ml = '[remove_postpatch_script]';

    $rc = pp_mysystem("rm -f $postpatch_script");
    if ($rc == 0) {
	loginfo("$ml script removed: $postpatch_script");
    }
    else {
	logerror("$ml could not remove script: $postpatch_script (status=$rc)");
	$rc = 0;
    }

    return($rc);
}


sub pp_remove_ftd_subs_file
{
    my ($patched_daisy_dir) = @_;

    my $rc = 0;

    my $ftd_subs_file = File::Spec->catdir($patched_daisy_dir, dsy_pathto_ftd_subs_file($patched_daisy_dir));
    if (-f $ftd_subs_file) {
	loginfo("Removing Daisy FTD subs file: $ftd_subs_file");
	$rc = pp_mysystem("rm -f $ftd_subs_file");
	if ($rc == 0) {
	    loginfo("[remove_ftd_sub_file] FTD subs file removed: $ftd_subs_file");
	}
	else {
	    logerror("[remove_ftd_sub_file] could not remove FTD subs file: $ftd_subs_file");
	}
    }

    return($rc);
}


#
# Get the system runlevel.
#
# Returns 0-6 on success, -1 on error
#
sub pp_get_runlevel
{
    my $runlevel = -1;

    my $ml ='[get_runlevel]';

    #
    # Output of runlevel is:
    #	(digit|letter) space digit newline
    # where the right most digit is the current runlevel
    #
    my $cmd = '/sbin/runlevel';
    if (open(my $pipe, q{-|}, $cmd)) {
	while (<$pipe>) {
	    $runlevel = $_;
	    chomp($runlevel);
	    $runlevel =~ s/^. (\d)$/$1/;
	}
	close($pipe) or warn "[$ml] could not close pipe to command: $cmd ($OS_ERROR)\n";;
    }
    else {
	showerror("$ml could not open pipe to command: $cmd");
    }

    return($runlevel);
}


#
# Set the system runlevel to value between 0 and 6.
#
# returns
#   1 on success
#   0 on error
#
sub pp_set_runlevel
{
    my ($new_runlevel) = @_;

    my $rc = 1;
    my $ml = '[set_runlevel]';

    if ($new_runlevel < 0 || $new_runlevel > 6) {
	logerror("$ml runlevel out of range: $new_runlevel");
	return(0);
    }

    system("/sbin/telinit $new_runlevel");
    if ($? == 0) {
	loginfo("$ml runlevel set to: $new_runlevel");

	# The "init" man page mentions that "init" waits 5 seconds
	# between each of two kills, and testing reveals we need
	# to wait a bit for the runlevel to change.
	sleep(30);
    }
    else {
	logerror("$ml could not set runlevel to: $new_runlevel");
	$rc = 0;
    }

    return($rc);
}


#
# Stop all Daisy processes.
#
# returns
#   1 on success
#   0 on error
#
sub pp_stop_daisy
{
    my $rc = 1;
    my $ml = '[stop_daisy]';

    # try to change the system runlevel
    if (pp_set_runlevel(4)) {
	loginfo("$ml runlevel of system set to: 4");
    }
    else {
	logerror("$ml could not set system runlevel to: 4");
	return(0);
    }

    # verify system runlevel is what we expect
    if (pp_get_runlevel() == 4) {
	loginf("$ml verified system runlvel is: 4");
    }
    else {
	logerror("$ml could not verify system runlevel is: 4");
	return(0);
    }

    # now kill off all daisy processes
    my $killemall_cmd = "/d/daisy/bin/killemall.pl";
    if (-f $killemall_cmd) {
	system("$killemall_cmd");
	if ($? == 0) {
	    loginfo("$ml Daisy processes killed");
	}
	else {
	    logerror("$ml could not kill Daisy processes");
	    return(0);
	}
    }
    else {
	logerror("$ml script does not exist: $killemall_cmd");
	return(0);
    }

    return($rc);
}


#
# Start Daisy.
#
# Returns 0 on success, -1 on error
#
sub pp_start_daisy
{
    my $rc = 0;

    if (pp_set_runlevel(3) != 0) {
	logerror("Can't set runlevel to 3");
	return(-1);
    }

    if (pp_get_runlevel() != 3) {
	logerror("Unable to return to runlevel 3");
	return(-1);
    }

    loginfo("Daisy restarted and system back to runlevel 3");

    return(0);
}


#
# Read shop code from Dove control file in default daisy database dir.
#
# Returns:
#   shop code as 8 digit string if found else empty string
#
sub pp_get_shopcode
{
    my ($patched_daisy_dir) = @_;

    my $shopcode = $EMPTY_STR;
    my $ml = '[get_shopcode]';

    if (-d $patched_daisy_dir) {
	my $dovectrl_file = File::Spec->catdir($patched_daisy_dir, $DOVE_CONTROL_FILENAME);

	if (! -f $dovectrl_file) {
	    logerror("$ml Dove control file does not exist: $dovectrl_file");
	    return($EMPTY_STR);
	}

	if (open(my $fh, '<', $dovectrl_file)) {
	    my $buffer = "";
	    my $bufsize = 50;
	    my $rc = sysread($fh, $buffer, $bufsize);

	    if (! defined($rc)) {
		logerror("$ml error reading Dove control file: $dovectrl_file ($!)");
		return($EMPTY_STR);
	    }
	    if ($rc == 0) {
		logerror("$ml could not read Dove control file: $dovectrl_file");
		return($EMPTY_STR);
	    }
	    if ($rc != $bufsize) {
		logerror("$ml short read from Dove control file: $dovectrl_file");
		logerror("$ml Expected read of $bufsize but got $rc");
		return($EMPTY_STR);
	    }

	    close($fh) or warn "$ml could not close dove control file: $dovectrl_file ($OS_ERROR)\n";

	    $shopcode = substr($buffer, 30, 8);
	}
	else {
	    logerror("$ml could not open Dove control file: $dovectrl_file");
	    return($EMPTY_STR);
	}
    }

    return($shopcode);
}


#
# Look up the shop code from "the list", ie list of shops that have purchased
# the FTD files.
#
# returns
#   1 if shopcode is on the list
#   0 if not
#
sub pp_lookup_shopcode
{
    my ($patched_daisy_dir, $shopcode) = @_;

    my $rc = 0;
    my $ml = '[lookup_shopcode]';
    my $ftd_subs_file = dsy_pathto_ftd_subs_file($patched_daisy_dir);

    #
    # Check to see if the shopcode is one of the QA shopcodes that
    # may be specified in the config file. 
    #
    if (@CONFIG_QA_SHOPCODE) {
	foreach (@CONFIG_QA_SHOPCODE) {
	    if (/$shopcode/) {
		loginfo("$ml shopcode matched shopcode from config file: $shopcode");
		loginfo("$ml FTD subscription allowed");
		return(1);
	    }
	}
    }

    if (! -f $ftd_subs_file) {
	logerror("$ml FTD subscribers file not found: $ftd_subs_file");
	loginfo("$ml FTD subscription allowed");
	return(1);
    } 

    if (open(my $fh, '<', $ftd_subs_file)) {
	while (<$fh>) {

	    # skip comments or blank lines
	    next if ( (/^(\s*)#/) || (/^(\s*)$/) );

	    # remove newline
	    chomp;

	    # remove leading and trailing SPACE chars
	    my $subscriber = pp_trim($_);

	    # finally do the comparison
	    if ($subscriber eq $shopcode) {
		loginfo("$ml shopcode matched: $shopcode");
		loginfo("$ml FTD subscription allowed");
		$rc = 1;
		last;
	    }
	}
	close($fh);
    }
    else {
	logerror("$ml could not open FTD subscribers file: $ftd_subs_file");
	loginfo("$ml FTD subscription allowed");
	return(1);
    }

    return($rc);
}


#
# Get the shop code, and lookup the shop code in the list of
# FTD subscribers.  If the shop code is not on the list, then
# remove the FTD files from the patched daisy dir and it's
# backup diretory.  BUT, save the new FTD files in case the
# customer negotiates effectively enough to let them keep
# the new FTD files.
#
# Returns
#   1 on success
#   0 on error
#
sub pp_resolve_ftd_files
{
    my ($patched_daisy_dir) = @_;

    my $rc = 1;
    my $ml = '[resolve_ftd_files]';

    #
    # if the command line option specified, leave the FTD files
    # installed without any further checks.
    #
    if ($CLO_INSTALL_FTD_FILES) {
	loginfo("$ml FTD files installation forced with \"--install-ftd-files\" option");
	return(1);
    }

    #
    # if the shop code can not be determined (which should not happend),
    # then allow FTD files.
    my $shopcode = pp_get_shopcode($patched_daisy_dir);
    if ($shopcode eq $EMPTY_STR) {
	logerror("$ml could not get shop code, keeping FTD files");
	return(1);
    }

    #
    # At this point, there may be old FTD files in the backup dir and
    # there may be new FTD files in the patched daisy dir.
    #
    # If the shopcode is not on "the list", then remove the new FTD files from
    # the patched daisy dir and the old FTD files from it's backup dir.

    if (pp_lookup_shopcode($patched_daisy_dir, $shopcode)) {
	loginfo("$ml shop is an FTD subscriber: $shopcode");
    }
    else {

	loginfo("$ml shop is NOT a FTD subscriber: $shopcode");

	if (pp_save_new_ftd_files($patched_daisy_dir)) {
	    loginfo("$ml new FTD files saved");
	}
	else {
	    logerror("$ml could not save new FTD files");
	}

	my $rm_status = pp_mysystem("rm -f $patched_daisy_dir/*.ftd");
	if ($rm_status == 0) {
	    loginfo("$ml removed new FTD files from: $patched_daisy_dir");
	}
	else {
	    logerror("$ml could not remove new FTD files from: $patched_daisy_dir (status=$rm_status)");
	    $rc = 0;
	}

	my $backup_dir = File::Spec->catdir($patched_daisy_dir, dsy_nameof_backup_dir());
	$rm_status = pp_mysystem("rm -f $backup_dir/*.ftd.prepatch.*");
	if ($rm_status == 0) {
	    loginfo("$ml removed old FTD files from: $backup_dir");
	}
	else {
	    logerror("$ml could not remove old FTD files from: $backup_dir (status=$rm_status)");
	    $rc = 0;
	}
    }

    # remove FTD subscriber file if it exists
    my $ftd_subs_file = dsy_pathto_ftd_subs_file($patched_daisy_dir);
    if (-e $ftd_subs_file) {
	my $rm_status = pp_mysystem("rm -f $ftd_subs_file");
	if ($rm_status == 0) {
	    loginfo("$ml FTD subscribers file removed: $ftd_subs_file");
	}
	else {
	    logerror("$ml could not remove FTD subscribers file: $ftd_subs_file");
	    $rc = 0;
	}
    }

    return($rc);
}


sub pp_resolve_hsp_files
{
    my ($patched_daisy_dir) = @_;

    my $rc = 1;
    my $ml ='[resolve_hsp_files]';

    my @hsp_backup_files = glob("$patched_daisy_dir/backups/hnot*.hsp.prepatch.*");

    if (scalar(@hsp_backup_files) == 0) {
	loginfo("$ml there are no hospital files in the backups directory");
	return($rc);
    }

    foreach my $hsp_file (@hsp_backup_files) {

	# isolate the filename
	my $hsp_filename = basename($hsp_file);

	# convert it back to original
	if ($hsp_filename =~ /(.*)\.prepatch\..*$/) {
	    $hsp_filename = $1;
	}

	$rc = pp_mysystem("cp $hsp_file $patched_daisy_dir/$hsp_filename");
	if ($rc == 0) {
	    loginfo("$ml restored pervious hospital note file: $hsp_filename");
	}
	else {
	    logerror("$ml could not restore hospital note file: $hsp_filename ($rc)");
	    $rc = 0;
	}
    }

    return($rc);
}


sub pp_save_new_ftd_files
{
    my ($patched_daisy_dir) = @_;

    my $rc = 1;
    my $ml = '[save_new_ftd_files]';

    # make sure '/d/edirectories' exists - make it if not
    my $florist_dir = dsy_pathto_florist_dir();
    if (-e $florist_dir) {
	loginfo("$ml directory for saved FTD files exists: $florist_dir");
    }
    else {
	loginfo("$ml directory for saved FTD files does not exist: $florist_dir");
	system("mkdir $florist_dir");
	if (-e $florist_dir) {
	    loginfo("$ml mkdir of directory for saved FTD files successful: $florist_dir");
	}
	else {
	    logerror("$ml could not make directory for saved FTD files: $florist_dir");
	    $rc = 0;
	}
    }

    # remove any old saved FTD files archives
    my $ftd_archives_re = $FTD_FILES_PREFIX . '*' . $FTD_FILES_SUFFIX;
    my $old_saved_ftd_arhives = File::Spec->catdir(dsy_pathto_florist_dir(), $ftd_archives_re);
    my $rm_status = pp_mysystem("rm -f $old_saved_ftd_arhives");
    if ($rm_status == 0) {
	loginfo("$ml old archives of saved FTD files removed: $old_saved_ftd_arhives");
    }
    else {
	logerror("$ml could not remove old archives of saved FTD files: $old_saved_ftd_arhives");
	$rc = 0;
    }

    # make tar archive from "$patched_daisy_dir/*.ftd"
    if ($rc) {
	my $ftd_files_archive = dsy_pathto_ftd_files_archive();
	chdir($patched_daisy_dir);
	system("tar cf $ftd_files_archive *.ftd");
	if (-e $ftd_files_archive) {
	    loginfo("$ml FTD files saved to: $ftd_files_archive");
	}
	else {
	    logerror("$ml could not save FTD files to: $ftd_files_archive");
	    $rc = 0;
	}
    }

    return($rc);
}


#
# return list of new binaries - list members will have the
# platform suffix removed.
#
# returns
#   list of new binaries, possibly empty list
#
sub pp_make_list_of_new_binaries
{
    my ($patched_daisy_dir) = @_;

    my @nu_binaries = ();
    my $ml = '[make_list_of_new_binaries]';

    if (opendir(my $dh, $patched_daisy_dir)) {
        foreach my $dirent (readdir($dh)) {
	    if ($dirent eq '.' || $dirent eq '..') {
		next;
	    }
	    foreach my $platform (@SUPPORTED_PLATFORMS) {
		if ($dirent =~ /.+-$platform/) {
		    push(@nu_binaries, $dirent);
		}
	    }
        }
        closedir($dh);
    }
    else {
	logerror("$ml could not open directory: $patched_daisy_dir");
    }

    return(@nu_binaries);
}


#
# resolve_binaries
#
# foreach new binary file
#   mv the platform specific file to the original file
#   fix perms
#
# returns
#   1 on success
#   0 otherwise
#
sub pp_resolve_binaries
{
    my ($patched_daisy_dir) = @_;

    my $rc = 1;
    my $ml ='[resolve_binaries]';

    my $os = pp_determine_os();

    my @nu_binaries = pp_make_list_of_new_binaries($patched_daisy_dir);

    #
    # make a list of binaries for this platform, remove
    # non-applicable platform binaries 
    #
    my @binaries = ();
    foreach my $nu_binary (@nu_binaries) {
	if ($nu_binary =~ /.+-$os/) {
	    push(@binaries, $nu_binary);
	}
	else {
	    pp_mysystem("rm $patched_daisy_dir/$nu_binary");
	}
    }

    #
    # move new binaries into place, set perms
    #
    foreach my $binary (@binaries) {

	my $realname = $binary;
	$realname =~ s/-$os//;

	my $mv_status = pp_mysystem("mv $patched_daisy_dir/$binary $patched_daisy_dir/$realname");
	if ($mv_status == 0) {
	    loginfo("$ml installed new binary: $realname");
	    system("chown tfsupport:daisy $patched_daisy_dir/$realname");
	    system("chmod 550 $patched_daisy_dir/$realname");
	}
	else {
	    logerror("$ml installation failed for: $realname ($mv_status)");
	    $rc = 0;
	}
    }

    return($rc);
}


#
# (1) If there are no FTD files in the backup dir, then the assumption is
#     that this daisy system does not subscribe to FTD.  So, remove
#     the FTD files from the current daisy db dir.
#
# (2) If there are "hospital note files" in the backup dir, that means
#     there were some in the daisy db dir before the patch was applied...
#     so the new empty files need to be removed from the current daisy
#     db dir and the old ones copied from the backup dir to the current
#     daisy db dir.
#
sub pp_postpatch_edir_update
{
    my ($patched_daisy_dir) = @_;

    my $rc = 1;
    my $ml = '[postpatch_edir_update]';

    if (pp_resolve_ftd_files($patched_daisy_dir)) {
	loginfo("$ml FTD files resolved");
    }
    else {
	logerror("$ml could not resolve FTD files");
	$rc = 0;
    }

    if (pp_resolve_hsp_files($patched_daisy_dir)) {
	loginfo("$ml HSP files resolved");
    }
    else {
	logerror("$ml could not resolve HSP files");
	$rc = 0;
    }

    return($rc);
}


# trim whitespace from start and end of string
sub pp_trim
{
    my ($string) = @_;

    $string =~ s/^\s+//;
    $string =~ s/\s+$//;

    return($string);
}

sub pp_determine_os
{
    my $os = $EMPTY_STR;
    my $ml = '[determine_os]';

    my $platform_file = '/etc/redhat-release';
    if (open(my $fh, '<', $platform_file)) {
	while(<$fh>) {

		# ES 7
		# Redhat Enterprise Linux Server 7
		if( (/(release)(\s+)(7)/)
		||  (/(CentOS)([[:print:]]+)(\s)(7)/) ) {
			$os = 'RHEL7';
			last;
		}

		# ES 6
		# Redhat Enterprise Linux Server 6
		if( (/(release)(\s+)(6)/)
		||  (/(CentOS)([[:print:]]+)(\s)(6)/) ) {
			$os = 'RHEL6';
			last;
		}

		# ES 5
		# Redhat Enterprise Linux Server 5
		if( (/(release)(\s+)(5)/)
		||  (/(CentOS)([[:print:]]+)(\s)(5)/) ) {
			$os = 'RHEL5';
			last;
		}
	}
	close($fh) or warn "$ml could not close config file: $platform_file ($OS_ERROR)\n";
    }
    else {
	logerror("$ml could not open file: $platform_file");
    }

    return($os);
}


#
# Subroutine to run system commands.  This logs an entry.
#
sub pp_mysystem
{
    my ($command) = @_;

    if ($CLO_DEBUG) {
	system("$command >> $LOGFILE 2>> $LOGFILE");
    }
    else {
	system($command);
    }
    my $returnval = WEXITSTATUS($?);
    if ($CLO_DEBUG) {
        loginfo("$command Returned: $returnval");
    }

    return ($returnval)
}



sub showinfo
{
    my ($message) = @_;

    print("$message\n");

    return(loginfo($message));
}

sub showerror
{
    my ($message) = @_;

    print("$message\n");

    return(logerror($message));
}

#
# Log an item to the logfile.
#
sub loginfo
{
    my ($message) = @_;

    chomp ($message);
    my $timestamp = strftime("%Y-%m-%d %H:%M:%S", localtime());

    if (open(my $fh, '>>', $LOGFILE)) {
	print {$fh} "$timestamp <I>: $message\n";
	close($fh);
    }

    return($EMPTY_STR);
}

#
# Log an error to the logfile.
#
sub logerror
{
    my ($message) = @_;

    chomp ($message);
    my $timestamp = strftime("%Y-%m-%d %H:%M:%S", localtime());

    if (open(my $fh, '>>', $LOGFILE)) {
	print {$fh} "$timestamp <E>: $message\n";
	close($fh);
    }

    return($EMPTY_STR);
}



__END__

=pod

=head1 NAME

postpatch.pl - apply custom changes after applying a patch

=head1 VERSION

This documenation refers to version: $Revision: 1.40 $


=head1 USAGE

postpatch.pl

postpatch.pl B<--version>

postpatch.pl B<--help>

postpatch.pl [options] daisy_dir time_stamp list_of_files\n");


=head1 OPTIONS

=over 4

=item B<--version>

Output the version number of the script and exit.

=item B<--help>

Output a short help message and exit.

=item B<--[no]edirupdate>

If this command line option is specfied,
in addition to performing the normal task of installing binary files and
removing the F<postpatch.pl> script itself,
the script manages installation of the FTD files and
the Daisy notes files.
The default is to perform the management of the FTD files.
To keep the script from managing FTD files, use the option negation.

=item B<--install-ftd-files>

If this command line option is specified,
the code which decides whether to allow FTD files will be skipped and
the FTD files will be installed.
The default is to not install the FTD files.

=item B<--debug>

If this command line option is specified,
the output of any call to the Perl F<system> function is
concatenated to the log file.
Also, the return value of the F<system> function is
concatenated to the log file.

=back


=head1 DESCRIPTION

The standard patch procedure for a Teleflora Daisy Point of Sales system
involves first running the F<applypatch.pl> script to apply a "patch",
i.e. a set of files contained in a compressed tar archive, and
then running the F<postpatch.pl> script which is customized as
appropriate to the contents of the "patch" file.
The "patch" file can contain updates to Daisy programs, new Daisy programs,
updates to data files, new data files, and/or
the files for the static florist directory update.
The F<applypatch.pl> script applies the patch to the default
Daisy database directory and any additional database directories
for sites with multiple database directories.

Usually, one of the files in the "patch" is this F<postpatch.pl> script, and
if it exists, it is executed by the F<applypatch.pl> script 
after the files from the "patch" are applied by F<applypatch.pl>
(really just copied into place after saving any old instances
of files in the patch in the backup directory).

Since the F<applypatch.pl> script applies the "patch" on each of
any existing Daisy database dirs, the F<postpatch.pl> script is run
on each existing Daisy database dir in turn.
Normally, the behavior of F<postpatch.pl> must be customized for each "patch".

=head2 Florist Directory Update

As mentioned, one use for the Daisy patch procedure is for
installing the periodic update to the Daisy florist directory files.
For this case, called the static florist directory update package,
the package consists of a tar archive containing the F<applypatch.pl> script and
the "patch" file, ie a compressed tar archive.
The "patch" file contains the set of files which constitute the florist directory update
along with the F<postpatch.pl> script.

Optionally, the Daisy static florist directory update package can contain
a file with a list of Teleflora shopcodes that are also FTD subscribers.
The name of this file is F<FTD_Subscribers.txt>.
If the FTD subscribers file exists in the update package,
when the F<postpatch.pl> script runs,
the script gets the shopcode of the Daisy system,
looks up that shopcode in the FTD subscribers file, and
if the shopcode is present in the FTD subscribes file,
the script allows the FTD files to stay on the system.
If the shopcode is NOT in the FTD subscribers file, then the
new FTD files are saved to the directory F</d/edirectories> in a tar archive
named "FTD_Files_YYYYMMDDhhmmss.tar" (where "YYYY" is the year,
"MM" is the month, "DD" is the day, "hh" is the hour, "mm" is the minute,
and "ss" is the seconds of the time the F<applypatch.pl> started execution),
the new FTD files are removed from the Daisy database directory, and
any old FTD files are removed from the Daisy backup directory (F</d/daisy/backups>).

If the shopcode of the system is defined in the
config file via the "qa-shopcode" variable,
then the check for deciding if the Daisy system is a FTD subscriber will pass and
the FTD files will be installed.


=head1 FILES

=over 4

=item B</d/daisy/config/applypatch.config>

The configuration file for the apply patch process.
Currently, the only variable that may be defined in the config file
is "qa-shopcode=nnnnnnnn" where "nnnnnnnn" represents a shopcode.
There may be zero or more of these lines in the config file.

=item B<FTD_Subscribers.txt>

This file contains a list of Teleflora shopcodes which are also FTD subscribers.
The file may be contained within a Daisy static florist directory update package.

If the shopcode of the system that F<postpatch.pl> is running on
does NOT appear in this file, then a message will be logged and
the FTD files will NOT be installed.

=item B</d/edirectories/FTD_Files_YYYYMMDDhhmmss.tar>

When applying a static florist directory update,
if the Daisy shop is not a FTD subsriber,
the new FTD files are saved to this file,
which is a tar archive.
The string "YYYYMMDDhhmmss" is a date stamp which represents
the start of execution of the update procedure.

=item B</d/daisy/log/postpatch.log>

The file that F<postpatch.pl> log messages are written to.

=item B</d/daisy/log/RTI-Patches.log>

The file that F<applypatch.pl> log messages are written to.

=item B</d/daisy/backups>

Each Daisy database directory has a directory named F<backups> which
is where old instances of files in the "patch" are copied before
being replaced with the new instances.

=back


=head1 DIAGNOSTICS

=over 4

=item Exit status 0

Successful completion or when the "--version" and "--help"
command line options are specified.

=item Exit status 1

If the path to the Daisy database directory to "post patch" is NOT
specified as the first argument on the command line.
If the time stamp is NOT
specified as the second argument on the command line.

=item Exit status 14

The Daisy statuc florist directory update failed.

=item Exit status 15

The post patch of Daisy programs failed.

=back


=head1 SEE ALSO

B<applypatch.pl>


=cut

